#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Module holding the inversion recovers Look-Locker quantification model."""
import numexpr as ne
import numpy as np
from pyqmri.models.template import BaseModel, constraints


class Model(BaseModel):
    """Inversion recovery Look-Locker model for MRI parameter quantification.

    This class holds a IRLL model for T1 quantification from
    complex MRI data. It realizes a forward application of the analytical
    signal expression and the partial derivatives with respesct to
    each parameter of interest, as required by the abstract methods in
    the BaseModel. The fitting target is the exponential term itself
    which is easier to fit than the corresponding timing constant.

    The rescale function applies a transformation and returns the expected
    T1 values in ms.

    The implemented signal model follows the work from Henderson et al. (1999)

    The model should only be used for radially acquired data!

    Parameters
    ----------
      par : dict
        A python dict containing the necessary information to
        setup the object. Needs to contain the sequence related parametrs,
        e.g. TR, TE, TI, to fully describe the acquisitio process

    Attributes
    ----------
      TR : float
        Repetition time of the IRLL sequence.
      fa : float
        Flip angle of the gradient echo sequence.
      tau : float
        Repetition time for one gradient echo read out.
      td : float
        Delay prio to first read-out point after inversion.
      Nproj : int
        Number of projections per bin
      Nproj_measrued : int
        Total number of projections measured
      uk_scale : list of float
        Scaling factors for each unknown to balance the partial derivatives.
      guess : numpy.array, None
        The initial guess. Needs to be set using "computeInitialGuess"
        prior to fitting.
      scale : float
        A scaling factor to balance the different exponential terms.

    """

    def __init__(self, par):
        super().__init__(par)

        self.TR = par['time_per_slice'] - \
            (par['tau'] * par['Nproj_measured'] + par['gradient_delay'])
        self.fa = par["flip_angle(s)"] * np.pi / 180

        self.tau = par["tau"]
        self.td = par["gradient_delay"]
        self.Nproj = par["Nproj"]
        self.Nproj_measured = par["Nproj_measured"]

        par["unknowns_TGV"] = 2
        par["unknowns_H1"] = 0
        par["unknowns"] = par["unknowns_TGV"]+par["unknowns_H1"]

        phi_corr = np.zeros_like(par["fa_corr"], dtype=self._DTYPE)
        phi_corr = np.real(
            self.fa) * np.real(par["fa_corr"]) + 1j *\
            np.imag(self.fa) * np.imag(par["fa_corr"])

        self._sin_phi = np.sin(phi_corr)
        self._cos_phi = np.cos(phi_corr)

        for j in range(par["unknowns"]):
            self.uk_scale.append(1)

        self.scale = 100

        self.constraints.append(constraints(0, 300, False))
        self.constraints.append(constraints(
            np.exp(-self.scale / 10) / self.uk_scale[1],
            np.exp(-self.scale / 5500) / self.uk_scale[1], True))

        self.guess = None

        self._ax = None

        self._M0_plot = None
        self._M0_plot_cor = None
        self._M0_plot_sag = None

        self._T1_plot = None
        self._T1_plot_cor = None
        self._T1_plot_sag = None

    def rescale(self, x):
        """Rescale the unknowns with the scaling factors.

        Rescales each unknown with the corresponding scaling factor and
        applies a transformation for the time constants of the exponentials,
        yielding a resulting T1 in milliseconds.

        Parameters
        ----------
          x : numpy.array
            The array of unknowns to be rescaled

        Returns
        -------
          numpy.array:
            The rescaled unknowns
        """
        tmp_x = np.copy(x)
        tmp_x[0] *= self.uk_scale[0]
        tmp_x[1] = -self.scale / np.log(tmp_x[1] * self.uk_scale[1])
        const = []
        for constrained in self.constraints:
            const.append(constrained.real)
        return {"data": tmp_x,
                "unknown_name": ["M0", "T1"],
                "real_valued": const}

    def _execute_forward_3D(self, x):
        S = np.zeros(
            (self.NScan,
             self.Nproj,
             self.NSlice,
             self.dimY,
             self.dimX),
            dtype=self._DTYPE)

        TR = self.TR
        tau = self.tau
        td = self.td
        sin_phi = self._sin_phi
        cos_phi = self._cos_phi
        N = self.Nproj_measured
        scale = self.scale
        Efit = x[1, ...] * self.uk_scale[1]
        Etau = Efit**(tau / scale)
        Etr = Efit**(TR / scale)
        Etd = Efit**(td / scale)
        M0 = x[0, ...]
        M0_sc = self.uk_scale[0]

        F = (1 - Etau) / (1 - Etau * cos_phi)
        Q = (-Etr * Etd * F * (-(Etau * cos_phi)**(N - 1) + 1) *
             cos_phi + Etr *
             Etd - 2 * Etd + 1) / (Etr * Etd * (Etau * cos_phi)**(N - 1) *
                                   cos_phi + 1)
        Q_F = Q - F

        def numexpeval_S(M0, M0_sc, sin_phi, cos_phi, n, Q_F, F, Etau):
            return ne.evaluate(
                "M0*M0_sc*((Etau*cos_phi)**(n - 1)*Q_F + F)*sin_phi")
        for i in range(self.NScan):
            for j in range(self.Nproj):
                n = i * self.Nproj + j + 1
                S[i, j, ...] = numexpeval_S(
                    M0, M0_sc, sin_phi, cos_phi, n, Q_F, F, Etau)

        return np.mean(S, axis=1, dtype=self._DTYPE).astype(self._DTYPE)

    def _execute_gradient_3D(self, x):
        grad = np.zeros(
            (2,
             self.NScan,
             self.Nproj,
             self.NSlice,
             self.dimY,
             self.dimX),
            dtype=self._DTYPE)
        M0_sc = self.uk_scale[0]
        TR = self.TR
        tau = self.tau
        td = self.td
        sin_phi = self._sin_phi
        cos_phi = self._cos_phi
        N = self.Nproj_measured
        scale = self.scale
        Efit = x[1, ...] * self.uk_scale[1]
        Etau = Efit**(tau / scale)
        Etr = Efit**(TR / scale)
        Etd = Efit**(td / scale)
        M0 = x[0, ...]
        M0_sc = self.uk_scale[0]

        F = (1 - Etau) / (1 - Etau * cos_phi)
        Q = (-Etr * Etd * (-Etau + 1) * (-(Etau * cos_phi)**(N - 1) + 1) *
             cos_phi / (-Etau * cos_phi + 1) + Etr * Etd - 2 * Etd + 1) / \
            (Etr * Etd * (Etau * cos_phi)**(N - 1) * cos_phi + 1)
        Q_F = Q - F
        tmp1 = (
            (
                - TR * Etr * Etd * (-Etau + 1)
                * (-(Etau * cos_phi)**(N - 1) + 1)
                * cos_phi / (x[1, ...] * scale * (-Etau * cos_phi + 1))
                + TR * Etr * Etd / (x[1, ...] * scale) - tau * Etr * Etau
                * Etd * (-Etau + 1)
                * (-(Etau * cos_phi)**(N - 1) + 1) * cos_phi**2
                / (x[1, ...] * scale * (-Etau * cos_phi + 1)**2)
                + tau * Etr * Etau * Etd * (-(Etau * cos_phi)**(N - 1) + 1)
                * cos_phi / (x[1, ...] * scale * (-Etau * cos_phi + 1))
                + tau * Etr * Etd * (Etau * cos_phi)**(N - 1)
                * (N - 1) * (-Etau + 1) * cos_phi
                / (x[1, ...] * scale * (-Etau * cos_phi + 1))
                - td * Etr * Etd * (-Etau + 1)
                * (-(Etau * cos_phi)**(N - 1) + 1) * cos_phi
                / (x[1, ...] * scale * (-Etau * cos_phi + 1))
                + td * Etr * Etd / (x[1, ...] * scale) - 2 *
                td * Etd / (x[1, ...] * scale)
            ) / (Etr * Etd * (Etau * cos_phi)**(N - 1) * cos_phi + 1)
            + (
                -TR * Etr * Etd * (Etau * cos_phi)**(N - 1) * cos_phi
                / (x[1, ...] * scale) - tau * Etr * Etd
                * (Etau * cos_phi)**(N - 1) * (N - 1)
                * cos_phi / (x[1, ...] * scale) - td * Etr * Etd
                * (Etau * cos_phi)**(N - 1) * cos_phi
                / (x[1, ...] * scale)
            ) * (
                -Etr * Etd * (-Etau + 1)
                * (-(Etau * cos_phi)**(N - 1) + 1) * cos_phi
                / (-Etau * cos_phi + 1) + Etr * Etd - 2 * Etd + 1
            ) / (
                Etr * Etd * (Etau * cos_phi)**(N - 1) * cos_phi + 1
            )**2 -
            tau * Etau * (-Etau + 1) * cos_phi
            / (x[1, ...] * scale * (-Etau * cos_phi + 1)**2)
            + tau * Etau / (x[1, ...] * scale * (-Etau * cos_phi + 1))
            )

        tmp2 = (
            tau * Etau * (-Etau + 1) * cos_phi
            / (x[1, ...] * scale * (-Etau * cos_phi + 1)**2)
            - tau * Etau / (x[1, ...] * scale * (-Etau * cos_phi + 1))
            )

        tmp3 = (
            -(-Etau + 1) / (-Etau * cos_phi + 1)
            + (
                - Etr * Etd * (-Etau + 1) * (-(Etau * cos_phi)**(N - 1) + 1)
                * cos_phi / (-Etau * cos_phi + 1) + Etr * Etd - 2 * Etd + 1
                ) / (
                    Etr * Etd * (Etau * cos_phi)**(N - 1) * cos_phi +
                    1
                    )
            ) / (x[1, ...] * scale)

        def numexpeval_M0(M0_sc, sin_phi, cos_phi, n, Q_F, F, Etau):
            return ne.evaluate(
                "M0_sc*((Etau*cos_phi)**(n - 1)*Q_F + F)*sin_phi")

        def numexpeval_T1(
                M0,
                M0_sc,
                Etau,
                cos_phi,
                sin_phi,
                n,
                tmp1,
                tmp2,
                tmp3,
                tau):
            return ne.evaluate(
                "M0*M0_sc*((Etau*cos_phi)**(n - 1)*tmp1 + "
                "tmp2 + tau*(Etau*cos_phi)**(n - 1)*(n - 1)*tmp3)*sin_phi")

        for i in range(self.NScan):
            for j in range(self.Nproj):
                n = i * self.Nproj + j + 1

                grad[0, i, j, ...] = numexpeval_M0(
                    M0_sc, sin_phi, cos_phi, n, Q_F, F, Etau)
                grad[1, i, j, ...] = numexpeval_T1(
                    M0, M0_sc, Etau, cos_phi, sin_phi, n,
                    tmp1, tmp2, tmp3, tau)

        return np.mean(grad, axis=2, dtype=self._DTYPE).astype(self._DTYPE)

    def computeInitialGuess(self, *args):
        """Initialize unknown array for the fitting.

        This function provides an initial guess for the fitting.

        Parameters
        ----------
          args : list of objects
            Serves as universal interface. No objects need to be passed
            here.
        """
        test_T1 = 800 * np.ones(
            (self.NSlice, self.dimY, self.dimX), dtype=self._DTYPE)
        test_M0 = 1e-3*np.ones(
            (self.NSlice, self.dimY, self.dimX), dtype=self._DTYPE)
        test_T1 = np.exp(-self.scale / (test_T1))

        self.guess = np.array(
            [test_M0 / self.uk_scale[0],
             test_T1 / self.uk_scale[1]], dtype=self._DTYPE)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Compute Kaiser-Bessel gridding kernel.

Original Skript by B. Hargreaves
Adapted for Python by O. Maier

function [kern,kbu] = calckbkernel(kwidth,overgridfactor,klength)
Function calculates the appropriate Kaiser-Bessel kernel
for gridding, using the approach of Jackson et al.
INPUT:
  width = kernel width in grid samples.
  overgridfactor = over-gridding factor.
  klength = kernel look-up-table length.
OUTPUT:
  kern = kernel values for klength values
  of u, uniformly spaced from 0 to kwidth/2.
  kbu = u values.
"""
import numpy as np


def calckbkernel(kwidth, overgridfactor, G, klength=32):
    """Precompute the Kaiser-Bessel Kerenl used for gridding.

    This function precomputes a Kaiser-Bessel kernel for a given width.
    To save memory, only half the kernel is computed due to its symmetry
    around zero.

    Parameters
    ----------
      kwidth : float
        The width of the kernel in Gridunits
      overgridfactor : float
        The overgridfactor compared to the desired image matrix.
      G : int
          The number of acquired data points per projection.
      klength : int, 32
        The number of discretization points of the kernel.

    Returns
    -------
        numpy.array
            The precomputed Kaiser-Bessel kernel
        numpy.array
            The fourier transformation of the kernel

    """
    if klength < 2:
        klength = 2
        print('Warning:  klength must be 2 or more - using 2.')

    a = overgridfactor
    w = kwidth
    # From Beatty et al.
    beta = np.pi * np.sqrt(w**2 / a**2 * (a - 0.5)**2 - 0.8)

    # Kernel radii - grid samples.
    u = np.linspace(0, np.floor(klength * w / 2), int(np.ceil(klength * w / 2))
                    ) / (np.floor(klength * w / 2)) * w / 2 / G

    kern = _kb(u, kwidth, beta, G)
    kern = kern / kern[u == 0]  # Normalize.

    ft_y = np.flip(kern)
    ft_y = np.concatenate((ft_y[:-1], kern))
    ft_y = np.pad(ft_y, int((G * klength - ft_y.size) / 2), 'constant')
    ft_y = np.abs(np.fft.fftshift(np.fft.ifft(
        np.fft.ifftshift(ft_y))) * ft_y.size)
    x = np.linspace(-int(G / (2 * a)), int(G / (2 * a)) - 1, int(G / (a)))

    ft_y = ft_y[(ft_y.size / 2 - x).astype(int)]
    h = np.sinc(x / (G * klength))**2

    kern_ft = (ft_y * h)
    kern_ft = (kern_ft / np.max((kern_ft)))

    return (kern, kern_ft)


def _kb(u, w, beta, G):
    if np.size(w) > 1:
        raise TypeError('w should be a single scalar value.')

    y = 0 * u  # Allocate space.
    uz = np.where(np.abs(u) <= w / (2 * G))			# Indices where u<w/2.

    if np.size(uz) > 0:			# Calculate y at indices uz.
        # Argument - see Jackson '91.
        x = beta * np.sqrt(1 - (2 * u[uz] * G / w)**2)
        y[uz] = G * np.i0(x) / w

    return y



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Module holding the general model for fitting."""
import configparser
import numpy as np
import sympy
from pyqmri.models.template import BaseModel, constraints


def _str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    if v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    raise ValueError('Boolean value expected.')


class Model(BaseModel):
    """Realization of a generative model based on sympy.

      This model can handel all kinds of sympy input in form of a config file.
      Partial derivatives of the model are automatically generated and a
      numpy compatible function is build from sumpy equations.

    Attributes
    ----------
      signaleq : sympy derived function
        The signal equation derived from sympy
      grad : list of functions
        Partial derivatives with respect to the unknowns
      rescalefun : list of functions
      Functions to rescale each parameter
      modelparams : list
        List of model parameters
      indphase : bool
        Flag to estimate the phase from a given image series.
        The phase is normed on the first image.
        If True, each image will be multiplied by the estimated phase in the
        forward and gradient evaluation.
      init_values : list of str
          Initial guess for each unknown
    """

    def __init__(self, par):

        super().__init__(par)
        config = configparser.ConfigParser()

        if not par["modelfile"].endswith('.ini'):
            par["modelfile"] += '.ini'
        try:
            with open(par["modelfile"], 'r') as f:
                config.read_file(f)
        except BaseException:
            print("Model file not readable or not found")
            raise BaseException
        finally:
            params = {}
            for key in config[par["modelname"]]:
                params[key] = config[par["modelname"]][key]

        modelpar = sympy.symbols(params["parameter"])
        unknowns = sympy.symbols(params["unknowns"])
        self._unknowns = unknowns

        par["unknowns_TGV"] = len(unknowns)
        par["unknowns_H1"] = 0
        par["unknowns"] = par["unknowns_TGV"]+par["unknowns_H1"]

        uk_scale = []
        for j in range(par["unknowns"]):
            uk_scale.append(sympy.symbols(str(unknowns[j])+"_sc"))
            params["signal"] = params["signal"].replace(
                str(unknowns[j]),
                "("+str(unknowns[j])+"*"+str(unknowns[j])+"_sc)")
            params["rescale"] = params["rescale"].replace(
                str(unknowns[j]),
                "("+str(unknowns[j])+"*"+str(unknowns[j])+"_sc)")
        signaleq = sympy.sympify(params["signal"])
        params["rescale"] = params["rescale"].split(",")

        self.grad = []
        self.rescalefun = []

        for uk, scalefuns in zip(unknowns, params["rescale"]):
            tmp_grad = sympy.diff(signaleq, uk)
            self.grad.append(sympy.lambdify(
                (modelpar, unknowns, uk_scale), tmp_grad))
            self.rescalefun.append(
                sympy.lambdify(
                    (modelpar, unknowns, uk_scale),
                    sympy.sympify(scalefuns)))

        self.signaleq = sympy.lambdify(
            (modelpar, unknowns, uk_scale), signaleq)

        self.modelparams = []
        for mypar in modelpar:
            tmp = par[str(mypar)]
            if np.isscalar(tmp) or tmp.shape == (1,):
                self.modelparams.append(tmp)
            elif (len(tmp.shape) >= 3 and
                  tmp.shape[-3:] == (self.NSlice, self.dimY, self.dimX)):
                self.modelparams.append(tmp)
            else:
                while len(tmp.shape) < 4:
                    tmp = tmp[..., None]
                self.modelparams.append(tmp)

        self.uk_scale = []
        for j in range(par["unknowns"]):
            self.uk_scale.append(1)

        box_constraints_low = params["box_constraints_lower"].split(",")
        box_constraints_up = params["box_constraints_upper"].split(",")
        real_const = params["real_value_constraints"].split(",")
        for uk in range(par["unknowns"]):
            self.constraints.append(
                constraints(
                    float(box_constraints_low[uk]),
                    float(box_constraints_up[uk]),
                    _str2bool(real_const[uk])))

        self.indphase = _str2bool(params["estimate_individual_phase"])

        self.init_values = params["guess"].split(",")

        self._plot = []
        self._phase = None
        self.guess = None

    def rescale(self, x):
        """Rescale the unknowns with the scaling factors.

        Rescales each unknown with the corresponding scaling factor and
        an optional transformation.

        Parameters
        ----------
          x : numpy.array
            The array of unknowns to be rescaled

        Returns
        -------
          numpy.array:
            The rescaled unknowns
        """
        tmp_x = np.copy(x)
        uk_name = []
        for j in range(x.shape[0]):
            tmp_x[j] = self.rescalefun[j](
                self.modelparams, x, self.uk_scale)
            uk_name.append(str(self._unknowns[j]))
        const = []
        for constrained in self.constraints:
            const.append(constrained.real)
        return {"data": tmp_x,
                "unknown_name": uk_name,
                "real_valued": const}

    def _execute_forward_3D(self, x):
        S = self.signaleq(self.modelparams, x, self.uk_scale)
        while len(S.shape) >= 5:
            S = np.squeeze(S, axis=0)
        if self.indphase is True:
            S *= self._phase
        S[~np.isfinite(S)] = 1e-20
        S = S.astype(dtype=self._DTYPE)
        return S

    def _execute_gradient_3D(self, x):
        modelgradient = []
        if self.indphase is True:
            for ukgrad in self.grad:
                modelgradient.append(
                    ukgrad(self.modelparams, x, self.uk_scale)*self._phase)
        else:
            for ukgrad in self.grad:
                modelgradient.append(
                    ukgrad(self.modelparams, x, self.uk_scale))
        modelgradient = np.array(modelgradient, dtype=self._DTYPE)
        while len(modelgradient.shape) >= 6:
            modelgradient = np.squeeze(modelgradient, axis=1)
        modelgradient[~np.isfinite(modelgradient)] = 1e-20
        return modelgradient

    def computeInitialGuess(self, *args):
        """Initialize unknown array for the fitting.

        This function provides an initial guess for the fitting, based
        on the values on the text file.

        Parameters
        ----------
          args : list of objects
            Assumes the image series at potition 0 and optionally computes
            a phase based on the difference between each image series minus
            the first image in the series. (Scan i minus Scan 0)
        """
        if self.indphase is True:
            self._phase = np.exp(1j*(np.angle(args[0])-np.angle(args[0][0])))
        x = np.ones((len(self.init_values),
                     self.NSlice, self.dimY, self.dimX), self._DTYPE)
        for j in range(len(self.init_values)):
            if "image" in self.init_values[j]:
                x[j] = args[0][int(self.init_values[j].split("_")[-1])]
            else:
                x[j] *= float(self.init_values[j])
        self.guess = x


def genDefaultModelfile():
    """Generate a default model config file.

    This method generates a default model file in the current project folder.
    This file can be modified or further models can be added.
    """
    config = configparser.ConfigParser()

    config["MonoExp"] = {}
    config["MonoExp"]["parameter"] = "TE"
    config["MonoExp"]["unknowns"] = "M0 A1"
    config["MonoExp"]["signal"] = "M0*exp(-TE*A1)"
    config["MonoExp"]["box_constraints_lower"] = "0,0"
    config["MonoExp"]["box_constraints_upper"] = "100,1"
    config["MonoExp"]["real_value_constraints"] = "False,True"
    config["MonoExp"]["guess"] = "1,0.01"
    config["MonoExp"]["rescale"] = "M0,1/A1"
    config["MonoExp"]["estimate_individual_phase"] = "False"

    config["VFA-E1"] = {}
    config["VFA-E1"]["parameter"] = "TR fa fa_corr"
    config["VFA-E1"]["unknowns"] = "M0 E_1"
    config["VFA-E1"]["signal"] = "M0*sin(fa*fa_corr)*"\
                                 "(1-E_1)/(1-E_1*cos(fa*fa_corr))"
    config["VFA-E1"]["box_constraints_lower"] = "0,0.9048"
    config["VFA-E1"]["box_constraints_upper"] = "10,0.99909"
    config["VFA-E1"]["real_value_constraints"] = "False,True"
    config["VFA-E1"]["guess"] = "1,0.99667"
    config["VFA-E1"]["rescale"] = "M0,-TR/log(E_1)"
    config["VFA-E1"]["estimate_individual_phase"] = "False"

    with open('models.ini', 'w') as configfile:
        config.write(configfile)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Module holding the template base class model."""
import itertools
from abc import ABC, abstractmethod
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec


class constraints:
    """Constraints for a parameter.

    This class holds min/max and real value constrains for a parameter. It
    supports updating these based on the current estimated sclaing between
    each partial derivative.

    Parameters
    ----------
      min_val : float, -numpy.inf
        The minimum value.
      max_val : float, numpy.inf
        The maximum value.
      real_const : bool, false
        Constrain to real values (true) or complex values (false).

    Attributes
    ----------
      constraints : list of pyqmri.models.template.constrains
        An empy list of constrains objects.
      NScan : int
        Number of scans (dynamics).
      NSlice : int
        Number of slices.
      dimX, dimY : int
        The image dimensions.
      figure : matplotlib.pyplot.figure, None
        The placeholder figure object
    """

    def __init__(self, min_val=-np.inf, max_val=np.inf,
                 real_const=False):
        self.min = min_val
        self.max = max_val
        self.real = real_const

    def update(self, scale):
        """Update the constrains based on current scaling factor.

        Parameters
        ----------
          scale : float
            The new scaling factor which should be used.
        """
        self.min = self.min / scale
        self.max = self.max / scale


class BaseModel(ABC):
    """Base model for MRI parameter quantification.

    This class holds the base model to derive other signal models from.
    It defines abstract a forward application of the analytical signal
    expression    and the partial derivatives with respesct to each parameter
    of interest.

    Parameters
    ----------
      par : dict
        A python dict containing the necessary information to
        setup the object. Needs to contain the sequence related parametrs,
        e.g. Number of Scans, Slices, and image dimension.

    Attributes
    ----------
      constraints : list of pyqmri.models.template.constrains
        An empy list of constrains objects.
      NScan : int
        Number of scans (dynamics).
      NSlice : int
        Number of slices.
      dimX, dimY : int
        The image dimensions.
    """

    def __init__(self, par):
        super().__init__()
        self.constraints = []
        self.uk_scale = []
        self.NScan = par["NScan"]
        self.NSlice = par["NSlice"]
        self.dimX = par["dimX"]
        self.dimY = par["dimY"]
        self._DTYPE = par["DTYPE"]
        self._DTYPE_real = par["DTYPE_real"]
        self._figure = None
        self._plot_trans = []
        self._plot_cor = []
        self._plot_sag = []

    def rescale(self, x):
        """Rescale the unknowns with the scaling factors.

        Rescales each unknown with the corresponding scaling factor.

        Parameters
        ----------
          x : numpy.array
            The array of unknowns to be rescaled

        Returns
        -------
          numpy.array:
            The rescaled unknowns
        """
        tmp_x = np.copy(x)
        uk_names = []
        for i in range(x.shape[0]):
            tmp_x[i] *= self.uk_scale[i]
            uk_names.append("Unkown_"+str(i))
        const = []
        for constrained in self.constraints:
            const.append(constrained.real)
        return {"data": tmp_x,
                "unknown_name": uk_names,
                "real_valued": const}

    def execute_forward(self, x, islice=None):
        """Execute the signal model from parameter to imagespace.

        This function exectues the given signal model to generate an image
        series, given estimated parameters.

        Parameters
        ----------
          x : numpy.array
            The array of quantitative parameters to be fitted
          islice : int, None
            Currently unused.
        """
        # if islice is None:
        return self._execute_forward_3D(x)

    def execute_gradient(self, x, islice=None):
        """Execute the partial derivatives of the signal model.

        This function exectues the partial derivatives with respect to each
        unknown parameter, based on the signal model.

        Parameters
        ----------
          x : numpy.array
            The array of quantitative parameters to be fitted
          islice : int, None
            Currently unused.
        """
        # if islice is None:
        return self._execute_gradient_3D(x)

    @abstractmethod
    def _execute_forward_3D(self, x):
        ...

    @abstractmethod
    def _execute_gradient_3D(self, x):
        ...

    def plot_unknowns(self, x):
        """Plot the unkowns in an interactive figure.

        This function can be used to plot intermediate results during the
        optimization process.

        Parameters
        ----------
          x : dict
            A Python dictionary containing the array of unknowns to be
            displayed, the associated names and real value constrains.
        """
        unknowns = self.rescale(x)
        NSlice = unknowns["data"].shape[1]
        numunknowns = unknowns["data"].shape[0]
        [dimZ, dimY, dimX] = unknowns["data"].shape[1:]

        if not self._figure:
            plot_dim_x = int(np.ceil(np.sqrt(numunknowns)))
            plot_dim_y = int(np.round(np.sqrt(numunknowns)))
            plt.ion()
            self._figure = plt.figure(figsize=(12, 6))
            self._figure.subplots_adjust(hspace=0.3, wspace=0)
            wd_ratio = np.tile([1, dimZ/dimX, 1 / 20, 1 / (5)], plot_dim_x)
            hd_ratio = np.tile([1, dimZ/dimY, 1 / (5)], plot_dim_y)
            gs = gridspec.GridSpec(
                3 * plot_dim_y, 4 * plot_dim_x,
                width_ratios=wd_ratio,
                height_ratios=hd_ratio,
                hspace=0,
                wspace=0)
            gs.tight_layout(self._figure)
            self._figure.patch.set_facecolor(plt.cm.viridis.colors[0])

            for i, j in itertools.product(
                    range(plot_dim_y),
                    range(plot_dim_x)):
                if i*plot_dim_x+j < numunknowns:
                    ax_trans = plt.subplot(gs[3*i, 4*j])
                    ax_sag = plt.subplot(gs[3*i, 4*j+1])
                    ax_cor = plt.subplot(gs[3*i+1, 4*j])
                    ax_bar = plt.subplot(gs[3*i:3*i+2, 4*j+2])
                    ax_trans.axis('off')
                    ax_sag.axis('off')
                    ax_cor.axis('off')

                    ax_trans.set_anchor('SE')
                    ax_sag.set_anchor('SW')
                    ax_cor.set_anchor('NE')

                    if unknowns["real_valued"][i*plot_dim_x+j]:
                        def mytrafo(x):
                            return np.real(x)
                    else:
                        def mytrafo(x):
                            return np.abs(x)

                    self._plot_trans.append(
                        ax_trans.imshow(
                            mytrafo(unknowns["data"][i*plot_dim_x+j,
                                                     int(NSlice / 2), ...])))
                    self._plot_cor.append(
                        ax_cor.imshow(
                            mytrafo(unknowns["data"][i*plot_dim_x+j,
                                                     ..., int(dimY / 2), :])))
                    self._plot_sag.append(
                        ax_sag.imshow(
                            mytrafo(unknowns["data"][i*plot_dim_x+j,
                                                     ..., int(dimX / 2)].T)))
                    ax_trans.set_title(
                        unknowns["unknown_name"][i*plot_dim_x+j],
                        color='white')

                    cbar = self._figure.colorbar(
                        self._plot_trans[-1], cax=ax_bar)
                    cbar.ax.tick_params(labelsize=12, colors='white')
                    for spine in cbar.ax.spines:
                        cbar.ax.spines[spine].set_color('white')
            plt.draw()
            plt.pause(1e-10)

        else:
            for j in range(numunknowns):
                if unknowns["real_valued"][j]:
                    def mytrafo(x):
                        return np.real(x)
                else:
                    def mytrafo(x):
                        return np.abs(x)
                self._plot_trans[j].set_data(
                    mytrafo(unknowns["data"][j,
                                             int(NSlice / 2), ...]))
                self._plot_cor[j].set_data(
                    mytrafo(unknowns["data"][j,
                                             ..., int(dimY / 2), :]))
                self._plot_sag[j].set_data(
                    mytrafo(unknowns["data"][j,
                                             ..., int(dimX / 2)].T))
                minval = mytrafo(unknowns["data"][j]).min()
                maxval = mytrafo(unknowns["data"][j]).max()
                self._plot_trans[j].set_clim([minval, maxval])
                self._plot_cor[j].set_clim([minval, maxval])
                self._plot_sag[j].set_clim([minval, maxval])
            plt.draw()
            plt.pause(1e-10)

    @abstractmethod
    def computeInitialGuess(self, *args):
        """Initialize unknown array for the fitting.

        This function provides an initial guess for the fitting.
        """
        ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Module holding the classes for different linear Operators."""
from abc import ABC, abstractmethod
import pyopencl.array as clarray
import numpy as np
from pyqmri.transforms import PyOpenCLnuFFT as CLnuFFT
import pyqmri.streaming as streaming


class Operator(ABC):
    """Abstract base class for linear Operators used in the optimization.

    This class serves as the base class for all linear operators used in
    the varous optimization algorithms. it requires to implement a forward
    and backward application in and out of place.

    Parameters
    ----------
      par : dict
        A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      prg : PyOpenCL.Program
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator.
      DTYPE : numpy.dtype, numpy.complex64
         Complex working precission.
      DTYPE_real : numpy.dtype, numpy.float32
          Real working precission.

    Attributes
    ----------
      NScan : int
        Number of total measurements (Scans)
      NC : int
        Number of complex coils
      NSlice : int
        Number ofSlices
      dimX : int
        X dimension of the parameter maps
      dimY : int
        Y dimension of the parameter maps
      N : int
        N number of samples per readout
      Nproj : int
        Number of rreadouts
      unknowns_TGV : int
        Number of unknowns which should be regularized with TGV. It is assumed
        that these occure first in the unknown vector. Currently at least 1
        TGV unknown is required.
      unknowns_H1 : int
        Number of unknowns which should be regularized with H1. It is assumed
        that these occure after all TGV unknowns in the unknown vector.
        Currently this number can be zero which implies that no H1
        regularization is used.
      unknowns : int
        The sum of TGV and H1 unknowns.
      ctx : list of PyOpenCL.Context
        The context for the PyOpenCL computations. If streamed operations are
        used a list of ctx is required. One for each computation device.
      queue : list of PyOpenCL.Queue
        The computation Queue for the PyOpenCL kernels. If streamed operations
        are used a list of queues is required. Four for each computation
        device.
      dz : float
        The ratio between the physical X,Y dimensions vs the Z dimension.
        This allows for anisotrpic regularization along the Z dimension.
      num_dev : int
        Number of compute devices
      NUFFT : PyQMRI.transforms.PyOpenCLnuFFT
        A PyOpenCLnuFFT object to perform forward and backword transformations
        from image to k-space and vice versa.
      prg : PyOpenCL.Program
        The PyOpenCL program containing all compiled kernels.
      self.DTYPE : numpy.dtype
        Complex working precission. Currently single precission only.
      self.DTYPE_real : numpy.dtype
        Real working precission. Currently single precission only.
    """

    def __init__(self, par, prg, DTYPE=np.complex64, DTYPE_real=np.float32):
        self.NSlice = par["NSlice"]
        self.NScan = par["NScan"]
        self.dimX = par["dimX"]
        self.dimY = par["dimY"]
        self.N = par["N"]
        self.NC = par["NC"]
        self.Nproj = par["Nproj"]
        self.ctx = par["ctx"]
        self.queue = par["queue"]
        self.unknowns_TGV = par["unknowns_TGV"]
        self.unknowns_H1 = par["unknowns_H1"]
        self.unknowns = par["unknowns"]
        self._dz = par["dz"]
        self.num_dev = len(par["num_dev"])
        self._tmp_result = []
        self.NUFFT = []
        self.prg = prg
        self.DTYPE = DTYPE
        self.DTYPE_real = DTYPE_real
        self.par_slices = self.NSlice
        self._overlap = 0

    @abstractmethod
    def fwd(self, out, inp, **kwargs):
        """Forward operator application in-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex measurement space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.

        Returns
        -------
          PyOpenCL.Event
            A PyOpenCL event to wait for.
        """
        ...

    @abstractmethod
    def adj(self, out, inp, **kwargs):
        """Adjoint operator application in-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex parameter space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex measurement space data which is used as input.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        ...

    @abstractmethod
    def fwdoop(self, inp, **kwargs):
        """Forward operator application out-of-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        ...

    @abstractmethod
    def adjoop(self, inp, **kwargs):
        """Adjoint operator application out-of-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex measurement space which is used as input.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        ...

    @staticmethod
    def MRIOperatorFactory(par,
                           prg,
                           DTYPE,
                           DTYPE_real,
                           trafo=False,
                           imagespace=False,
                           SMS=False,
                           streamed=False):
        """MRI forward/adjoint operator factory method.

        Parameters
        ----------
          par : dict A python dict containing the necessary information to
            setup the object. Needs to contain the number of slices (NSlice),
            number of scans (NScan), image dimensions (dimX, dimY), number of
            coils (NC), sampling points (N) and read outs (NProj)
            a PyOpenCL queue (queue) and the complex coil
            sensitivities (C).
          prg : PyOpenCL.Program
            The PyOpenCL.Program object containing the necessary kernels to
            execute the linear Operator.
          DTYPE : numpy.dtype, numpy.complex64
             Complex working precission.
          DTYPE_real : numpy.dtype, numpy.float32
            Real working precission.
          trafo : bool, false
            Select between radial (True) or cartesian FFT (false).
          imagespace : bool, false
            Select between fitting in imagespace (True) or k-space (false).
          SMS : bool, false
            Select between simulatneous multi-slice reconstruction or standard.
          streamed : bool, false
            Use standard reconstruction (false) or streaming of memory blocks
            to the compute device (true). Only use this if data does not
            fit in one block.

        Returns
        -------
          PyQMRI.Operator
            A specialized instance of a PyQMRI.Operator to perform forward
            and ajoint operations for fitting.
          PyQMRI.NUFFT
            An instance of the used (nu-)FFT if k-space fitting is performed,
            None otherwise.
        """
        if streamed:
            if imagespace:
                op = OperatorImagespaceStreamed(
                    par, prg,
                    DTYPE=DTYPE,
                    DTYPE_real=DTYPE_real)
                FT = None
            else:
                if SMS:
                    op = OperatorKspaceSMSStreamed(
                        par,
                        prg,
                        DTYPE=DTYPE,
                        DTYPE_real=DTYPE_real)
                else:
                    op = OperatorKspaceStreamed(
                        par,
                        prg,
                        trafo=trafo,
                        DTYPE=DTYPE,
                        DTYPE_real=DTYPE_real)
                FT = op.NUFFT
        else:
            if imagespace:
                op = OperatorImagespace(
                    par, prg[0],
                    DTYPE=DTYPE,
                    DTYPE_real=DTYPE_real)
                FT = None
            else:
                if SMS:
                    op = OperatorKspaceSMS(
                        par,
                        prg[0],
                        DTYPE=DTYPE,
                        DTYPE_real=DTYPE_real)
                else:
                    op = OperatorKspace(
                        par,
                        prg[0],
                        trafo=trafo,
                        DTYPE=DTYPE,
                        DTYPE_real=DTYPE_real)
                FT = op.NUFFT
        return op, FT

    @staticmethod
    def GradientOperatorFactory(par,
                                prg,
                                DTYPE,
                                DTYPE_real,
                                streamed=False):
        """Gradient forward/adjoint operator factory method.

        Parameters
        ----------
          par : dict A python dict containing the necessary information to
            setup the object. Needs to contain the number of slices (NSlice),
            number of scans (NScan), image dimensions (dimX, dimY), number of
            coils (NC), sampling points (N) and read outs (NProj)
            a PyOpenCL queue (queue) and the complex coil
            sensitivities (C).
          prg : PyOpenCL.Program
            The PyOpenCL.Program object containing the necessary kernels to
            execute the linear Operator.
          DTYPE : numpy.dtype, numpy.complex64
             Complex working precission.
          DTYPE_real : numpy.dtype, numpy.float32
            Real working precission.
          streamed : bool, false
            Use standard reconstruction (false) or streaming of memory blocks
            to the compute device (true). Only use this if data does not
            fit in one block.

        Returns
        -------
          PyQMRI.Operator
            A specialized instance of a PyQMRI.Operator to perform forward
            and ajoint gradient calculations.
        """
        if streamed:
            op = OperatorFiniteGradientStreamed(par,
                                                prg,
                                                DTYPE,
                                                DTYPE_real)
        else:
            op = OperatorFiniteGradient(par,
                                        prg[0],
                                        DTYPE,
                                        DTYPE_real)
        return op

    @staticmethod
    def SymGradientOperatorFactory(par,
                                   prg,
                                   DTYPE,
                                   DTYPE_real,
                                   streamed=False):
        """Symmetrized Gradient forward/adjoint operator factory method.

        Parameters
        ----------
          par : dict A python dict containing the necessary information to
            setup the object. Needs to contain the number of slices (NSlice),
            number of scans (NScan), image dimensions (dimX, dimY), number of
            coils (NC), sampling points (N) and read outs (NProj)
            a PyOpenCL queue (queue) and the complex coil
            sensitivities (C).
          prg : PyOpenCL.Program
            The PyOpenCL.Program object containing the necessary kernels to
            execute the linear Operator.
          DTYPE : numpy.dtype, numpy.complex64
             Complex working precission.
          DTYPE_real : numpy.dtype, numpy.float32
            Real working precission.
          streamed : bool, false
            Use standard reconstruction (false) or streaming of memory blocks
            to the compute device (true). Only use this if data does not
            fit in one block.

        Returns
        -------
          PyQMRI.Operator
            A specialized instance of a PyQMRI.Operator to perform forward
            and ajoint symmetriced gradient calculations.
        """
        if streamed:
            op = OperatorFiniteSymGradientStreamed(par,
                                                   prg,
                                                   DTYPE,
                                                   DTYPE_real)
        else:
            op = OperatorFiniteSymGradient(par,
                                           prg[0],
                                           DTYPE,
                                           DTYPE_real)
        return op

    def _defineoperator(self,
                        functions,
                        outp,
                        inp,
                        reverse_dir=False,
                        posofnorm=None,
                        slices=None):
        if slices is None:
            slices = self.NSlice
        return streaming.Stream(
            functions,
            outp,
            inp,
            self.par_slices,
            self._overlap,
            slices,
            self.queue,
            self.num_dev,
            reverse_dir,
            posofnorm,
            DTYPE=self.DTYPE)


class OperatorImagespace(Operator):
    """Imagespace based Operator.

    This class serves as linear operator between parameter and imagespace.

    Use this operator if you want to perform complex parameter fitting from
    complex image space data without the need of performing FFTs.

    Parameters
    ----------
      par : dict A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      prg : PyOpenCL.Program
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator.
      DTYPE : numpy.dtype, numpy.complex64
         Complex working precission.
      DTYPE_real : numpy.dtype, numpy.float32
          Real working precission.

    Attributes
    ----------
    ctx : PyOpenCL.Context
      The context for the PyOpenCL computations.
    queue : PyOpenCL.Queue
      The computation Queue for the PyOpenCL kernels.
    """

    def __init__(self, par, prg, DTYPE=np.complex64, DTYPE_real=np.float32):
        super().__init__(par, prg, DTYPE, DTYPE_real)
        self.queue = self.queue[0]
        self.ctx = self.ctx[0]

    def fwd(self, out, inp, **kwargs):
        """Forward operator application in-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex measurement space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event
            A PyOpenCL event to wait for.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        return self.prg.operator_fwd_imagespace(
            self.queue, (self.NSlice, self.dimY, self.dimX), None,
            out.data, inp[0].data, inp[2].data,
            np.int32(self.NScan),
            np.int32(self.unknowns),
            wait_for=inp[0].events + out.events + wait_for)

    def fwdoop(self, inp, **kwargs):
        """Forward operator application out-of-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        tmp_result = clarray.empty(
            self.queue, (self.NScan, self.NSlice, self.dimY, self.dimX),
            self.DTYPE, "C")
        tmp_result.add_event(self.prg.operator_fwd_imagespace(
            self.queue, (self.NSlice, self.dimY, self.dimX), None,
            tmp_result.data, inp[0].data, inp[2].data,
            np.int32(self.NScan),
            np.int32(self.unknowns),
            wait_for=inp[0].events + wait_for))
        return tmp_result

    def adj(self, out, inp, **kwargs):
        """Adjoint operator application in-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex parameter space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex measurement space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        return self.prg.operator_ad_imagespace(
            out.queue, (self.NSlice, self.dimY, self.dimX), None,
            out.data, inp[0].data, inp[2].data,
            np.int32(self.NScan),
            np.int32(self.unknowns),
            wait_for=wait_for + inp[0].events + out.events)

    def adjoop(self, inp, **kwargs):
        """Adjoint operator application out-of-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex measurement space which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        out = clarray.empty(
            self.queue, (self.unknowns, self.NSlice, self.dimY, self.dimX),
            dtype=self.DTYPE)
        self.prg.operator_ad_imagespace(
            out.queue, (self.NSlice, self.dimY, self.dimX), None,
            out.data, inp[0].data, inp[2].data,
            np.int32(self.NScan),
            np.int32(self.unknowns),
            wait_for=wait_for + inp[0].events + out.events).wait()
        return out

    def adjKyk1(self, out, inp, **kwargs):
        """Apply the linear operator from image space to parameter space.

        This method fully implements the combined linear operator
        consisting of the data part as well as the TGV regularization part.

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex parameter space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex image space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        return self.prg.update_Kyk1_imagespace(
            self.queue, (self.NSlice, self.dimY, self.dimX), None,
            out.data, inp[0].data, inp[3].data, inp[1].data,
            np.int32(self.NScan),
            inp[4].data,
            np.int32(self.unknowns),
            self.DTYPE_real(self._dz),
            wait_for=(inp[0].events + out.events
                      + inp[1].events + wait_for))


class OperatorKspace(Operator):
    """k-Space based Operator.

    This class serves as linear operator between parameter and k-space.

    Use this operator if you want to perform complex parameter fitting from
    complex k-space data. The type of fft is defined through the NUFFT object.
    The NUFFT object can also be used for simple Cartesian FFTs.

    Parameters
    ----------
      par : dict A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      prg : PyOpenCL.Program
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator.
      DTYPE : numpy.dtype, numpy.complex64
        Complex working precission.
      DTYPE_real : numpy.dtype, numpy.float32
        Real working precission.
      trafo : bool, true
        Switch between cartesian (false) and non-cartesian FFT (True, default).

    Attributes
    ----------
    ctx : PyOpenCL.Context
      The context for the PyOpenCL computations.
    queue : PyOpenCL.Queue
      The computation Queue for the PyOpenCL kernels.
    NUFFT : PyQMRI.PyOpenCLnuFFT
      The (nu) FFT used for fitting.
    """

    def __init__(self, par, prg, DTYPE=np.complex64,
                 DTYPE_real=np.float32, trafo=True):
        super().__init__(par, prg, DTYPE, DTYPE_real)
        self.queue = self.queue[0]
        self.ctx = self.ctx[0]
        self._tmp_result = clarray.empty(
            self.queue, (self.NScan, self.NC,
                         self.NSlice, self.dimY, self.dimX),
            self.DTYPE, "C")
        if not trafo:
            self.Nproj = self.dimY
            self.N = self.dimX
        self.NUFFT = CLnuFFT.create(self.ctx,
                                    self.queue,
                                    par,
                                    radial=trafo,
                                    DTYPE=DTYPE,
                                    DTYPE_real=DTYPE_real)

    def fwd(self, out, inp, **kwargs):
        """Forward operator application in-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex measurement space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event
            A PyOpenCL event to wait for.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        self._tmp_result.add_event(
            self.prg.operator_fwd(
                self.queue,
                (self.NSlice, self.dimY, self.dimX),
                None,
                self._tmp_result.data, inp[0].data,
                inp[1].data,
                inp[2].data, np.int32(self.NC),
                np.int32(self.NScan),
                np.int32(self.unknowns),
                wait_for=(self._tmp_result.events + inp[0].events
                          + wait_for)))
        return self.NUFFT.FFT(
            out,
            self._tmp_result,
            wait_for=wait_for +
            self._tmp_result.events)

    def fwdoop(self, inp, **kwargs):
        """Forward operator application out-of-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        self._tmp_result.add_event(
            self.prg.operator_fwd(
                self.queue,
                (self.NSlice, self.dimY, self.dimX),
                None,
                self._tmp_result.data, inp[0].data,
                inp[1].data,
                inp[2].data, np.int32(self.NC),
                np.int32(self.NScan),
                np.int32(self.unknowns),
                wait_for=(self._tmp_result.events + inp[0].events
                          + wait_for)))
        tmp_sino = clarray.empty(
            self.queue,
            (self.NScan, self.NC, self.NSlice, self.Nproj, self.N),
            self.DTYPE, "C")
        tmp_sino.add_event(
            self.NUFFT.FFT(tmp_sino, self._tmp_result))
        return tmp_sino

    def adj(self, out, inp, **kwargs):
        """Adjoint operator application in-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex parameter space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex measurement space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        self._tmp_result.add_event(
            self.NUFFT.FFTH(
                self._tmp_result, inp[0], wait_for=(wait_for
                                                    + inp[0].events)))
        return self.prg.operator_ad(
            self.queue, (self.NSlice, self.dimY, self.dimX), None,
            out.data, self._tmp_result.data, inp[1].data,
            inp[2].data, np.int32(self.NC),
            np.int32(self.NScan),
            np.int32(self.unknowns),
            wait_for=self._tmp_result.events + out.events)

    def adjoop(self, inp, **kwargs):
        """Adjoint operator application out-of-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex measurement space which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        self._tmp_result.add_event(
            self.NUFFT.FFTH(
                self._tmp_result, inp[0], wait_for=(wait_for
                                                    + inp[0].events)))
        out = clarray.empty(
            self.queue, (self.unknowns, self.NSlice, self.dimY, self.dimX),
            dtype=self.DTYPE)
        self.prg.operator_ad(
            out.queue, (self.NSlice, self.dimY, self.dimX), None,
            out.data, self._tmp_result.data, inp[1].data,
            inp[2].data, np.int32(self.NC),
            np.int32(self.NScan),
            np.int32(self.unknowns),
            wait_for=(self._tmp_result.events
                      + out.events)).wait()
        return out

    def adjKyk1(self, out, inp, **kwargs):
        """Apply the linear operator from parameter space to k-space.

        This method fully implements the combined linear operator
        consisting of the data part as well as the TGV regularization part.

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex parameter space data which is used as input.
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        self._tmp_result.add_event(
            self.NUFFT.FFTH(
                self._tmp_result, inp[0], wait_for=(wait_for
                                                    + inp[0].events)))
        return self.prg.update_Kyk1(
            self.queue, (self.NSlice, self.dimY, self.dimX), None,
            out.data, self._tmp_result.data, inp[2].data,
            inp[3].data, inp[1].data, np.int32(self.NC),
            np.int32(self.NScan),
            inp[4].data,
            np.int32(self.unknowns), self.DTYPE_real(self._dz),
            wait_for=(self._tmp_result.events +
                      out.events + inp[1].events))


class OperatorKspaceSMS(Operator):
    """k-Space based Operator for SMS reconstruction.

    This class serves as linear operator between parameter and k-space.
    It implements simultaneous-multi-slice (SMS) reconstruction.

    Use this operator if you want to perform complex parameter fitting from
    complex k-space data measured with SMS. Currently only Cartesian FFTs are
    supported.

    Parameters
    ----------
      par : dict A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      prg : PyOpenCL.Program
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator.
      DTYPE : numpy.dtype, numpy.complex64
        Complex working precission.
      DTYPE_real : numpy.dtype, numpy.float32
        Real working precission.

    Attributes
    ----------
      packs : int
        Number of SMS packs.
      ctx : PyOpenCL.Context
        The context for the PyOpenCL computations.
      queue : PyOpenCL.Queue
        The computation Queue for the PyOpenCL kernels.
      NUFFT : PyQMRI.PyOpenCLnuFFT
        The (nu) FFT used for fitting.
    """

    def __init__(self, par, prg, DTYPE=np.complex64,
                 DTYPE_real=np.float32):
        super().__init__(par, prg, DTYPE, DTYPE_real)
        self.queue = self.queue[0]
        self.ctx = self.ctx[0]
        self.packs = par["packs"]*par["numofpacks"]
        self._tmp_result = clarray.empty(
            self.queue, (self.NScan, self.NC,
                         self.NSlice, self.dimY, self.dimX),
            self.DTYPE, "C")

        self.Nproj = self.dimY
        self.N = self.dimX
        self.NUFFT = CLnuFFT.create(self.ctx,
                                    self.queue,
                                    par,
                                    radial=False,
                                    SMS=True,
                                    DTYPE=DTYPE,
                                    DTYPE_real=DTYPE_real)

    def fwd(self, out, inp, **kwargs):
        """Forward operator application in-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex measurement space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event
            A PyOpenCL event to wait for.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        self._tmp_result.add_event(
            self.prg.operator_fwd(
                self.queue,
                (self.NSlice, self.dimY, self.dimX),
                None,
                self._tmp_result.data, inp[0].data,
                inp[1].data,
                inp[2].data, np.int32(self.NC),
                np.int32(self.NScan),
                np.int32(self.unknowns),
                wait_for=(self._tmp_result.events + inp[0].events
                          + wait_for)))
        return self.NUFFT.FFT(
            out,
            self._tmp_result,
            wait_for=self._tmp_result.events + out.events)

    def fwdoop(self, inp, **kwargs):
        """Forward operator application out-of-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        self._tmp_result.add_event(
            self.prg.operator_fwd(
                self.queue,
                (self.NSlice, self.dimY, self.dimX),
                None,
                self._tmp_result.data, inp[0].data,
                inp[1].data,
                inp[2].data, np.int32(self.NC),
                np.int32(self.NScan),
                np.int32(self.unknowns),
                wait_for=(self._tmp_result.events + inp[0].events
                          + wait_for)))
        tmp_sino = clarray.empty(
            self.queue,
            (self.NScan, self.NC, self.packs, self.Nproj, self.N),
            self.DTYPE, "C")
        tmp_sino.add_event(
            self.NUFFT.FFT(tmp_sino, self._tmp_result))
        return tmp_sino

    def adj(self, out, inp, **kwargs):
        """Adjoint operator application in-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex parameter space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex measurement space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        self._tmp_result.add_event(
            self.NUFFT.FFTH(
                self._tmp_result, inp[0], wait_for=(wait_for
                                                    + inp[0].events)))
        return self.prg.operator_ad(
            self.queue, (self.NSlice, self.dimY, self.dimX), None,
            out.data, self._tmp_result.data, inp[1].data,
            inp[2].data, np.int32(self.NC),
            np.int32(self.NScan),
            np.int32(self.unknowns),
            wait_for=self._tmp_result.events + out.events)

    def adjoop(self, inp, **kwargs):
        """Adjoint operator application out-of-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex measurement space which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        self._tmp_result.add_event(
            self.NUFFT.FFTH(
                self._tmp_result, inp[0], wait_for=(wait_for
                                                    + inp[0].events)))
        out = clarray.empty(
            self.queue, (self.unknowns, self.NSlice, self.dimY, self.dimX),
            dtype=self.DTYPE)
        self.prg.operator_ad(
            out.queue, (self.NSlice, self.dimY, self.dimX), None,
            out.data, self._tmp_result.data, inp[1].data,
            inp[2].data, np.int32(self.NC),
            np.int32(self.NScan),
            np.int32(self.unknowns),
            wait_for=self._tmp_result.events + out.events).wait()
        return out

    def adjKyk1(self, out, inp, **kwargs):
        """Apply the linear operator from parameter space to k-space.

        This method fully implements the combined linear operator
        consisting of the data part as well as the TGV regularization part.

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex parameter space data which is used as input.
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        self._tmp_result.add_event(
            self.NUFFT.FFTH(
                self._tmp_result, inp[0], wait_for=(wait_for
                                                    + inp[0].events)))
        return self.prg.update_Kyk1(
            self.queue, (self.NSlice, self.dimY, self.dimX), None,
            out.data, self._tmp_result.data, inp[2].data,
            inp[3].data, inp[1].data, np.int32(self.NC),
            np.int32(self.NScan),
            inp[4].data,
            np.int32(self.unknowns), self.DTYPE_real(self._dz),
            wait_for=(self._tmp_result.events +
                      out.events + inp[1].events))


class OperatorImagespaceStreamed(Operator):
    """The streamed version of the Imagespace based Operator.

    This class serves as linear operator between parameter and imagespace.
    All calculations are performed in a streamed fashion.

    Use this operator if you want to perform complex parameter fitting from
    complex image space data without the need of performing FFTs.
    In contrast to non-streaming classes no out of place operations
    are implemented.

    Parameters
    ----------
      par : dict A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      prg : PyOpenCL.Program
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator.
      DTYPE : numpy.dtype, numpy.complex64
        Complex working precission.
      DTYPE_real : numpy.dtype, numpy.float32
        Real working precission.

    Attributes
    ----------
      overlap : int
        Number of slices that overlap between adjacent blocks.
      par_slices : int
        Number of slices per streamed block
      fwdstr : PyQMRI.Stream
        The streaming object to perform the forward evaluation
      adjstr : PyQMRI.Stream
        The streaming object to perform the adjoint evaluation
      adjstrKyk1 : PyQMRI.Stream
        The streaming object to perform the adjoint evaluation including z1
        of the algorithm.
      unknown_shape : tuple of int
        Size of the parameter maps
      data_shape : tuple of int
        Size of the data
    """

    def __init__(self, par, prg, DTYPE=np.complex64, DTYPE_real=np.float32):
        super().__init__(par, prg, DTYPE, DTYPE_real)
        par["overlap"] = 1
        self._overlap = par["overlap"]
        self.par_slices = par["par_slices"]
        self.unknown_shape = (self.NSlice, self.unknowns, self.dimY, self.dimX)
        coil_shape = []
        model_grad_shape = (self.NSlice, self.unknowns,
                            self.NScan, self.dimY, self.dimX)
        self.data_shape = (self.NSlice, self.NScan, self.dimY, self.dimX)
        grad_shape = self.unknown_shape + (4,)

        self.fwdstr = self._defineoperator(
            [self._fwdstreamed],
            [self.data_shape],
            [[self.unknown_shape,
              coil_shape,
              model_grad_shape]])

        self.adjstrKyk1 = self._defineoperator(
            [self._adjstreamedKyk1],
            [self.unknown_shape],
            [[self.data_shape,
              grad_shape,
              coil_shape,
              model_grad_shape,
              self.unknown_shape]])

        self.adjstr = self._defineoperator(
            [self._adjstreamed],
            [self.unknown_shape],
            [[self.data_shape,
              coil_shape,
              model_grad_shape]])

    def fwd(self, out, inp, **kwargs):
        """Forward operator application in-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex measurement space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event
            A PyOpenCL event to wait for.
        """
        self.fwdstr.eval(out, inp)

    def fwdoop(self, inp, **kwargs):
        """Forward operator application out-of-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        tmp_result = np.zeros(self.data_shape, dtype=self.DTYPE)
        self.fwdstr.eval([tmp_result], inp)
        return tmp_result

    def adj(self, out, inp, **kwargs):
        """Adjoint operator application in-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex parameter space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex measurement space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        self.adjstr.eval(out, inp)

    def adjoop(self, inp, **kwargs):
        """Adjoint operator application out-of-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex measurement space which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        tmp_result = np.zeros(self.unknown_shape, dtype=self.DTYPE)
        self.adjstr.eval([tmp_result], inp)
        return tmp_result

    def adjKyk1(self, out, inp):
        """Apply the linear operator from parameter space to image space.

        This method fully implements the combined linear operator
        consisting of the data part as well as the TGV regularization part.

        Parameters
        ----------
          out : numpy.Array
            The complex parameter space data which is used as input.
          inp : numpy.Array
            The complex parameter space data which is used as input.
        """
        self.adjstrKyk1.eval(out, inp)

    def _fwdstreamed(self, outp, inp, par=None, idx=0, idxq=0,
                     bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        return (self.prg[idx].operator_fwd_imagespace(
            self.queue[4*idx+idxq],
            (self.par_slices+self._overlap, self.dimY, self.dimX), None,
            outp.data, inp[0].data,
            inp[2].data,
            np.int32(self.NScan), np.int32(self.unknowns),
            wait_for=outp.events+inp[0].events+inp[2].events+wait_for))

    def _adjstreamedKyk1(self, outp, inp, par=None, idx=0, idxq=0,
                         bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        return self.prg[idx].update_Kyk1_imagespace(
            self.queue[4*idx+idxq],
            (self.par_slices+self._overlap, self.dimY, self.dimX), None,
            outp.data, inp[0].data,
            inp[3].data,
            inp[1].data,
            np.int32(self.NScan),
            par[0][idx].data, np.int32(self.unknowns),
            np.int32(bound_cond), self.DTYPE_real(self._dz),
            wait_for=(outp.events+inp[0].events+inp[1].events +
                      inp[3].events+wait_for))

    def _adjstreamed(self, outp, inp, par=None, idx=0, idxq=0,
                     bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        return self.prg[idx].operator_ad_imagespace(
            self.queue[4*idx+idxq],
            (self.par_slices+self._overlap, self.dimY, self.dimX), None,
            outp.data, inp[0].data,
            inp[2].data,
            np.int32(self.NScan), np.int32(self.unknowns),
            wait_for=(outp.events+inp[0].events +
                      inp[2].events+wait_for))


class OperatorKspaceStreamed(Operator):
    """The streamed version of the k-space based Operator.

    This class serves as linear operator between parameter and k-space.
    All calculations are performed in a streamed fashion.

    Use this operator if you want to perform complex parameter fitting from
    complex k-space data without the need of performing FFTs.
    In contrast to non-streaming classes no out of place operations
    are implemented.

    Parameters
    ----------
      par : dict A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      prg : PyOpenCL.Program
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator.
      DTYPE : numpy.dtype, numpy.complex64
        Complex working precission.
      DTYPE_real : numpy.dtype, numpy.float32
        Real working precission.
      trafo : bool, true
        Switch between cartesian (false) and non-cartesian FFT (True, default).

    Attributes
    ----------
      overlap : int
        Number of slices that overlap between adjacent blocks.
      par_slices : int
        Number of slices per streamed block.
      fwdstr : PyQMRI.Stream
        The streaming object to perform the forward evaluation.
      adjstr : PyQMRI.Stream
        The streaming object to perform the adjoint evaluation.
      adjstrKyk1 : PyQMRI.Stream
        The streaming object to perform the adjoint evaluation including z1
        of the algorithm.
      NUFFT : list of PyQMRI.transforms.PyOpenCLnuFFT
        A list of NUFFT objects. One for each context.
      FTstr : PyQMRI.Stream
        A streamed version of the used (non-uniform) FFT, applied forward.
      unknown_shape : tuple of int
        Size of the parameter maps
      data_shape : tuple of int
        Size of the data
    """

    def __init__(self, par, prg,
                 DTYPE=np.complex64, DTYPE_real=np.float32, trafo=True):
        super().__init__(par, prg, DTYPE, DTYPE_real)
        self._overlap = par["overlap"]
        self.par_slices = par["par_slices"]
        if not trafo:
            self.Nproj = self.dimY
            self.N = self.dimX
        for j in range(self.num_dev):
            for i in range(2):
                self._tmp_result.append(
                    clarray.empty(
                        self.queue[4*j+i],
                        (self.par_slices+self._overlap, self.NScan,
                         self.NC, self.dimY, self.dimX),
                        self.DTYPE, "C"))
                self.NUFFT.append(
                    CLnuFFT.create(self.ctx[j],
                                   self.queue[4*j+i], par,
                                   radial=trafo,
                                   streamed=True,
                                   DTYPE=DTYPE,
                                   DTYPE_real=DTYPE_real))

        self.unknown_shape = (self.NSlice, self.unknowns, self.dimY, self.dimX)
        coil_shape = (self.NSlice, self.NC, self.dimY, self.dimX)
        model_grad_shape = (self.NSlice, self.unknowns,
                            self.NScan, self.dimY, self.dimX)
        self.data_shape = (self.NSlice, self.NScan, self.NC, self.Nproj,
                           self.N)
        trans_shape = (self.NSlice, self.NScan,
                       self.NC, self.dimY, self.dimX)
        grad_shape = self.unknown_shape + (4,)

        self.fwdstr = self._defineoperator(
            [self._fwdstreamed],
            [self.data_shape],
            [[self.unknown_shape,
              coil_shape,
              model_grad_shape]])

        self.adjstrKyk1 = self._defineoperator(
            [self._adjstreamedKyk1],
            [self.unknown_shape],
            [[self.data_shape,
              grad_shape,
              coil_shape,
              model_grad_shape,
              self.unknown_shape]])

        self.adjstr = self._defineoperator(
            [self._adjstreamed],
            [self.unknown_shape],
            [[self.data_shape,
              coil_shape,
              model_grad_shape]])

        self.FTstr = self._defineoperator(
            [self._FT],
            [self.data_shape],
            [[trans_shape]])

    def fwd(self, out, inp, **kwargs):
        """Forward operator application in-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex measurement space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event
            A PyOpenCL event to wait for.
        """
        self.fwdstr.eval(out, inp)

    def fwdoop(self, inp, **kwargs):
        """Forward operator application out-of-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        tmp_result = np.zeros(self.data_shape, dtype=self.DTYPE)
        self.fwdstr.eval([tmp_result], inp)
        return tmp_result

    def adj(self, out, inp, **kwargs):
        """Adjoint operator application in-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex parameter space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex measurement space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        self.adjstr.eval(out, inp)

    def adjoop(self, inp, **kwargs):
        """Adjoint operator application out-of-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex measurement space which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        tmp_result = np.zeros(self.unknown_shape, dtype=self.DTYPE)
        self.adjstr.eval([tmp_result], inp)
        return tmp_result

    def adjKyk1(self, out, inp):
        """Apply the linear operator from parameter space to k-space.

        This method fully implements the combined linear operator
        consisting of the data part as well as the TGV regularization part.

        Parameters
        ----------
          out : numpy.Array
            The complex parameter space data which is used as input.
          inp : numpy.Array
            The complex parameter space data which is used as input.
        """
        self.adjstrKyk1.eval(out, inp)

    def _fwdstreamed(self, outp, inp, par=None, idx=0, idxq=0,
                     bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        self._tmp_result[2*idx+idxq].add_event(self.prg[idx].operator_fwd(
            self.queue[4*idx+idxq],
            (self.par_slices+self._overlap, self.dimY, self.dimX), None,
            self._tmp_result[2*idx+idxq].data, inp[0].data,
            inp[1].data,
            inp[2].data,
            np.int32(self.NC),
            np.int32(self.NScan), np.int32(self.unknowns),
            wait_for=(self._tmp_result[2*idx+idxq].events +
                      inp[0].events+wait_for)))
        return self.NUFFT[2*idx+idxq].FFT(
            outp, self._tmp_result[2*idx+idxq],
            wait_for=outp.events+wait_for+self._tmp_result[2*idx+idxq].events)

    def _adjstreamedKyk1(self, outp, inp, par=None, idx=0, idxq=0,
                         bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        self._tmp_result[2*idx+idxq].add_event(
            self.NUFFT[2*idx+idxq].FFTH(
                self._tmp_result[2*idx+idxq], inp[0],
                wait_for=(wait_for+inp[0].events +
                          self._tmp_result[2*idx+idxq].events)))
        return self.prg[idx].update_Kyk1(
            self.queue[4*idx+idxq],
            (self.par_slices+self._overlap, self.dimY, self.dimX), None,
            outp.data, self._tmp_result[2*idx+idxq].data,
            inp[2].data,
            inp[3].data,
            inp[1].data, np.int32(self.NC), np.int32(self.NScan),
            par[0][idx].data, np.int32(self.unknowns),
            np.int32(bound_cond), self.DTYPE_real(self._dz),
            wait_for=(
                self._tmp_result[2*idx+idxq].events +
                outp.events+inp[1].events +
                inp[2].events + inp[3].events + wait_for))

    def _adjstreamed(self, outp, inp, par=None, idx=0, idxq=0,
                     bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        self._tmp_result[2*idx+idxq].add_event(
            self.NUFFT[2*idx+idxq].FFTH(
                self._tmp_result[2*idx+idxq], inp[0],
                wait_for=(wait_for+inp[0].events +
                          self._tmp_result[2*idx+idxq].events)))
        return self.prg[idx].operator_ad(
            self.queue[4*idx+idxq],
            (self.par_slices+self._overlap, self.dimY, self.dimX), None,
            outp.data, self._tmp_result[2*idx+idxq].data,
            inp[1].data,
            inp[2].data,
            np.int32(self.NC),
            np.int32(self.NScan), np.int32(self.unknowns),
            wait_for=(self._tmp_result[2*idx+idxq].events +
                      inp[1].events+inp[2].events+wait_for))

    def _FT(self, outp, inp, par=None, idx=0, idxq=0,
            bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        return self.NUFFT[2*idx+idxq].FFT(outp, inp[0])


class OperatorKspaceSMSStreamed(Operator):
    """The streamed version of the k-space based SMS Operator.

    This class serves as linear operator between parameter and k-space.
    It implements simultaneous-multi-slice (SMS) reconstruction.

    All calculations are performed in a streamed fashion.

    Use this operator if you want to perform complex parameter fitting from
    complex k-space data measured with SMS. Currently only Cartesian FFTs are
    supported.

    Parameters
    ----------
      par : dict A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      prg : PyOpenCL.Program
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator.
      DTYPE : numpy.dtype, numpy.complex64
        Complex working precission.
      DTYPE_real : numpy.dtype, numpy.float32
        Real working precission.


    Attributes
    ----------
      overlap : int
        Number of slices that overlap between adjacent blocks.
      par_slices : int
        Number of slices per streamed block
      packs : int
        Number of packs to stream
      fwdstr : PyQMRI.Stream
        The streaming object to perform the forward evaluation
      adjstr : PyQMRI.Stream
        The streaming object to perform the adjoint evaluation
      NUFFT : list of PyQMRI.transforms.PyOpenCLnuFFT
        A list of NUFFT objects. One for each context.
      FTstr : PyQMRI.Stream
        A streamed version of the used (non-uniform) FFT, applied forward.
      FTHstr : PyQMRI.Stream
        A streamed version of the used (non-uniform) FFT, applied adjoint.
      updateKyk1SMSstreamed
      dat_trans_axes : list of int
        Order in which the data needs to be transformed during the SMS
        reconstruction and streaming.
    """

    def __init__(self, par, prg, DTYPE=np.complex64,
                 DTYPE_real=np.float32):
        super().__init__(par, prg, DTYPE, DTYPE_real)
        self._overlap = par["overlap"]
        self.par_slices = par["par_slices"]
        self.packs = par["packs"]*par["numofpacks"]

        self.Nproj = self.dimY
        self.N = self.dimX

        for j in range(self.num_dev):
            for i in range(2):
                self._tmp_result.append(
                    clarray.empty(
                        self.queue[4*j+i],
                        (self.par_slices+self._overlap, self.NScan,
                         self.NC, self.dimY, self.dimX),
                        self.DTYPE, "C"))
                self.NUFFT.append(
                    CLnuFFT.create(self.ctx[j],
                                   self.queue[4*j+i], par,
                                   radial=False,
                                   SMS=True,
                                   streamed=True,
                                   DTYPE=DTYPE,
                                   DTYPE_real=DTYPE_real))

        unknown_shape = (self.NSlice, self.unknowns, self.dimY, self.dimX)
        coil_shape = (self.NSlice, self.NC, self.dimY, self.dimX)
        model_grad_shape = (self.NSlice, self.unknowns,
                            self.NScan, self.dimY, self.dimX)
        data_shape = (self.NSlice, self.NScan, self.NC, self.dimY, self.dimX)
        data_shape_T = (self.NScan, self.NC, self.packs,
                        self.dimY, self.dimX)
        trans_shape_T = (self.NScan,
                         self.NC, self.NSlice, self.dimY, self.dimX)
        grad_shape = unknown_shape + (4,)
        self.dat_trans_axes = [2, 0, 1, 3, 4]

        self.fwdstr = self._defineoperator(
            [self._fwdstreamed],
            [data_shape],
            [[unknown_shape,
              coil_shape,
              model_grad_shape]])

        self.adjstr = self._defineoperator(
            [self._adjstreamed],
            [unknown_shape],
            [[data_shape,
              coil_shape,
              model_grad_shape]])

        self.FTstr = self._defineoperatorSMS(
            [self._FT],
            [data_shape_T],
            [[trans_shape_T]])
        self.FTHstr = self._defineoperatorSMS(
            [self._FTH],
            [trans_shape_T],
            [[data_shape_T]])

        self._tmp_fft1 = np.zeros((self.NSlice, self.NScan, self.NC,
                                   self.dimY, self.dimX),
                                  dtype=self.DTYPE)
        self._tmp_fft2 = np.zeros((self.NScan, self.NC, self.NSlice,
                                   self.dimY, self.dimX),
                                  dtype=self.DTYPE)
        self._tmp_transformed = np.zeros((self.NScan, self.NC, self.packs,
                                          self.dimY, self.dimX),
                                         dtype=self.DTYPE)
        self._tmp_Kyk1 = np.zeros(unknown_shape,
                                  dtype=self.DTYPE)

        self._updateKyk1SMSStreamed = self._defineoperator(
            [self._updateKyk1SMS],
            [unknown_shape],
            [[unknown_shape,
              grad_shape,
              unknown_shape]],
            reverse_dir=True,
            posofnorm=[True])

    def fwd(self, out, inp, **kwargs):
        """Forward operator application in-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex measurement space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event
            A PyOpenCL event to wait for.
        """
        self.fwdstr.eval([self._tmp_fft1], inp)
        self._tmp_fft2 = np.require(
            np.transpose(
                self._tmp_fft1, (1, 2, 0, 3, 4)),
            requirements='C')
        self.FTstr.eval(
            [self._tmp_transformed],
            [[self._tmp_fft2]])
        out[0][...] = np.copy(np.require(
            np.transpose(
                self._tmp_transformed,
                self.dat_trans_axes),
            requirements='C'))

    def fwdoop(self, inp, **kwargs):
        """Forward operator application out-of-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        self.fwdstr.eval([self._tmp_fft1], inp)
        self._tmp_fft2 = np.require(
            np.transpose(
                self._tmp_fft1, (1, 2, 0, 3, 4)),
            requirements='C')
        self.FTstr.eval(
            [self._tmp_transformed],
            [[self._tmp_fft2]])
        return np.require(
            np.transpose(
                self._tmp_transformed,
                self.dat_trans_axes),
            requirements='C')

    def adj(self, out, inp, **kwargs):
        """Adjoint operator application in-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        Parameters
        ----------
          out : numpy.Array
            The complex parameter space data which is used as input.
          inp : numpy.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          tupel of floats:
            The lhs and rhs for the line search of the primal-dual algorithm.
        """
        self._tmp_transformed = np.require(
            np.transpose(
                inp[0][0], (1, 2, 0, 3, 4)),
            requirements='C')
        self.FTHstr.eval(
            [self._tmp_fft2],
            [[self._tmp_transformed]])
        self._tmp_fft1 = np.require(
            np.transpose(
                self._tmp_fft2, self.dat_trans_axes),
            requirements='C')
        self.adjstr.eval(out, [[self._tmp_fft1]+inp[0][1:]])

    def adjoop(self, inp, **kwargs):
        """Adjoint operator application out-of-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex measurement space which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        self._tmp_transformed = np.require(
            np.transpose(
                inp[0][0], (1, 2, 0, 3, 4)),
            requirements='C')
        self.FTHstr.eval(
            [self._tmp_fft2],
            [[self._tmp_transformed]])
        self._tmp_fft1 = np.require(
            np.transpose(
                self._tmp_fft2, self.dat_trans_axes),
            requirements='C')
        self.adjstr.eval([self._tmp_Kyk1], [[self._tmp_fft1]+inp[0][1:]])
        return self._tmp_Kyk1

    def adjKyk1(self, out, inp, **kwargs):
        """Apply the linear operator from parameter space to k-space.

        This method fully implements the combined linear operator
        consisting of the data part as well as the TGV regularization part.

        Parameters
        ----------
          out : numpy.Array
            The complex parameter space data which is used as input.
          inp : numpy.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          tupel of floats:
            The lhs and rhs for the line search of the primal-dual algorithm.
        """
        self._tmp_transformed = np.require(
            np.transpose(
                inp[0][0], (1, 2, 0, 3, 4)),
            requirements='C')
        self.FTHstr.eval(
            [self._tmp_fft2],
            [[self._tmp_transformed]])
        self._tmp_fft1 = np.require(
            np.transpose(
                self._tmp_fft2, self.dat_trans_axes),
            requirements='C')
        self.adjstr.eval([self._tmp_Kyk1], [[self._tmp_fft1]+inp[0][2:-1]])
        return self._updateKyk1SMSStreamed.evalwithnorm(
            out,
            [[self._tmp_Kyk1]+[inp[0][1]]+[inp[0][-1]]], kwargs["par"])

    def _fwdstreamed(self, outp, inp, par=None, idx=0, idxq=0,
                     bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        return self.prg[idx].operator_fwd(
            self.queue[4*idx+idxq],
            (self.par_slices+self._overlap, self.dimY, self.dimX), None,
            outp.data, inp[0].data,
            inp[1].data,
            inp[2].data,
            np.int32(self.NC),
            np.int32(self.NScan), np.int32(self.unknowns),
            wait_for=(outp.events+inp[0].events +
                      inp[1].events+inp[2].events+wait_for))

    def _adjstreamed(self, outp, inp, par=None, idx=0, idxq=0,
                     bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        return self.prg[idx].operator_ad(
            self.queue[4*idx+idxq],
            (self.par_slices+self._overlap, self.dimY, self.dimX), None,
            outp.data, inp[0].data,
            inp[1].data,
            inp[2].data,
            np.int32(self.NC), np.int32(self.NScan),
            np.int32(self.unknowns),
            wait_for=(
                inp[0].events +
                outp.events+inp[1].events +
                inp[2].events + wait_for))

    def _FT(self, outp, inp, par=None, idx=0, idxq=0,
            bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        return self.NUFFT[2*idx+idxq].FFT(outp, inp[0])

    def _FTH(self, outp, inp, par=None, idx=0, idxq=0,
             bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        return self.NUFFT[2*idx+idxq].FFTH(outp, inp[0])

    def _updateKyk1SMS(self, outp, inp, par=None, idx=0, idxq=0,
                       bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        return self.prg[idx].update_Kyk1SMS(
            self.queue[4*idx+idxq],
            (self.par_slices+self._overlap, self.dimY, self.dimX), None,
            outp.data, inp[0].data,
            inp[1].data,
            par[0][idx].data, np.int32(self.unknowns),
            np.int32(bound_cond), self.DTYPE_real(self._dz),
            wait_for=(
                inp[0].events +
                outp.events+inp[1].events +
                wait_for))

    def _defineoperatorSMS(self,
                           functions,
                           outp,
                           inp,
                           reverse_dir=False,
                           posofnorm=None):
        return streaming.Stream(
            functions,
            outp,
            inp,
            1,
            0,
            self.NScan,
            self.queue,
            self.num_dev,
            reverse_dir,
            posofnorm,
            DTYPE=self.DTYPE)


class OperatorFiniteGradient(Operator):
    """Gradient operator.

    This class implements the finite difference gradient operation and
    the adjoint (negative divergence).

    Parameters
    ----------
      par : dict A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      prg : PyOpenCL.Program
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator.
      DTYPE : numpy.dtype, numpy.complex64
        Complex working precission.
      DTYPE_real : numpy.dtype, numpy.float32
        Real working precission.

    Attributes
    ----------
      ctx : PyOpenCL.Context
        The context for the PyOpenCL computations.
      queue : PyOpenCL.Queue
        The computation Queue for the PyOpenCL kernels.
      ratio : list of PyOpenCL.Array
        Ratio between the different unknowns
    """

    def __init__(self, par, prg, DTYPE=np.complex64, DTYPE_real=np.float32):
        super().__init__(par, prg, DTYPE, DTYPE_real)
        self.queue = self.queue[0]
        self.ctx = self.ctx[0]
        self.ratio = clarray.to_device(
            self.queue,
            (par["weights"]).astype(
                     dtype=self.DTYPE_real))
        self._weights = par["weights"]

    def fwd(self, out, inp, **kwargs):
        """Forward operator application in-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event
            A PyOpenCL event to wait for.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        return self.prg.gradient(
            self.queue, inp.shape[1:], None, out.data, inp.data,
            np.int32(self.unknowns),
            self.ratio.data, self.DTYPE_real(self._dz),
            wait_for=out.events + inp.events + wait_for)

    def fwdoop(self, inp, **kwargs):
        """Forward operator application out-of-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        tmp_result = clarray.empty(
            self.queue, (self.unknowns,
                         self.NSlice, self.dimY, self.dimX, 4),
            self.DTYPE, "C")
        tmp_result.add_event(self.prg.gradient(
            self.queue, inp.shape[1:], None, tmp_result.data, inp.data,
            np.int32(self.unknowns),
            self.ratio.data, self.DTYPE_real(self._dz),
            wait_for=tmp_result.events + inp.events + wait_for))
        return tmp_result

    def adj(self, out, inp, **kwargs):
        """Adjoint operator application in-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex parameter space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex measurement space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        return self.prg.divergence(
            self.queue, inp.shape[1:-1], None, out.data, inp.data,
            np.int32(self.unknowns), self.ratio.data,
            self.DTYPE_real(self._dz),
            wait_for=out.events + inp.events + wait_for)

    def adjoop(self, inp, **kwargs):
        """Adjoint operator application out-of-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex measurement space which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        tmp_result = clarray.empty(
            self.queue, (self.unknowns, self.NSlice, self.dimY, self.dimX),
            self.DTYPE, "C")
        tmp_result.add_event(self.prg.divergence(
            self.queue, inp.shape[1:-1], None, tmp_result.data, inp.data,
            np.int32(self.unknowns), self.ratio.data,
            self.DTYPE_real(self._dz),
            wait_for=tmp_result.events + inp.events + wait_for))
        return tmp_result


class OperatorFiniteSymGradient(Operator):
    """Symmetrized gradient operator.

    This class implements the finite difference symmetrized gradient
    operation and the adjoint (negative symmetrized divergence).

    Parameters
    ----------
      par : dict A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      prg : PyOpenCL.Program
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator.
      DTYPE : numpy.dtype, numpy.complex64
        Complex working precission.
      DTYPE_real : numpy.dtype, numpy.float32
        Real working precission.

    Attributes
    ----------
      ctx : PyOpenCL.Context
        The context for the PyOpenCL computations.
      queue : PyOpenCL.Queue
        The computation Queue for the PyOpenCL kernels.
      ratio : list of PyOpenCL.Array
        Ratio between the different unknowns
    """

    def __init__(self, par, prg, DTYPE=np.complex64, DTYPE_real=np.float32):
        super().__init__(par, prg, DTYPE, DTYPE_real)
        self.queue = self.queue[0]
        self.ctx = self.ctx[0]
        self.ratio = clarray.to_device(
            self.queue,
            (par["weights"]).astype(
                     dtype=self.DTYPE_real))
        self._weights = par["weights"]

    def fwd(self, out, inp, **kwargs):
        """Forward operator application in-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event
            A PyOpenCL event to wait for.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        return self.prg.sym_grad(
            self.queue, inp.shape[1:-1], None, out.data, inp.data,
            np.int32(self.unknowns_TGV),
            self.ratio.data,
            self.DTYPE_real(self._dz),
            wait_for=out.events + inp.events + wait_for)

    def fwdoop(self, inp, **kwargs):
        """Forward operator application out-of-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        tmp_result = clarray.empty(
            self.queue, (self.unknowns,
                         self.NSlice, self.dimY, self.dimX, 8),
            self.DTYPE, "C")
        tmp_result.add_event(self.prg.sym_grad(
            self.queue, inp.shape[1:-1], None, tmp_result.data, inp.data,
            np.int32(self.unknowns_TGV),
            self.ratio.data,
            self.DTYPE_real(self._dz),
            wait_for=tmp_result.events + inp.events + wait_for))
        return tmp_result

    def adj(self, out, inp, **kwargs):
        """Adjoint operator application in-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex parameter space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex measurement space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        return self.prg.sym_divergence(
            self.queue, inp.shape[1:-1], None, out.data, inp.data,
            np.int32(self.unknowns_TGV),
            self.ratio.data,
            self.DTYPE_real(self._dz),
            wait_for=out.events + inp.events + wait_for)

    def adjoop(self, inp, **kwargs):
        """Adjoint operator application out-of-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex measurement space which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        if "wait_for" in kwargs.keys():
            wait_for = kwargs["wait_for"]
        else:
            wait_for = []
        tmp_result = clarray.empty(
            self.queue, (self.unknowns,
                         self.NSlice, self.dimY, self.dimX, 4),
            self.DTYPE, "C")
        tmp_result.add_event(self.prg.sym_divergence(
            self.queue, inp.shape[1:-1], None, tmp_result.data, inp.data,
            np.int32(self.unknowns_TGV),
            self.ratio.data,
            self.DTYPE_real(self._dz),
            wait_for=tmp_result.events + inp.events + wait_for))
        return tmp_result


class OperatorFiniteGradientStreamed(Operator):
    """Streamed gradient operator.

    This class implements the finite difference gradient
    operation and the adjoint (negative divergence).

    Parameters
    ----------
      par : dict A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      prg : PyOpenCL.Program
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator.
      DTYPE : numpy.dtype, numpy.complex64
        Complex working precission.
      DTYPE_real : numpy.dtype, numpy.float32
        Real working precission.

    Attributes
    ----------
      ctx : PyOpenCL.Context
        The context for the PyOpenCL computations.
      queue : PyOpenCL.Queue
        The computation Queue for the PyOpenCL kernels.
      par_slices : int
        Slices to parallel transfer to the compute device.
      ratio : list of PyOpenCL.Array
        Ratio between the different unknowns
    """

    def __init__(self, par, prg, DTYPE=np.complex64, DTYPE_real=np.float32):
        super().__init__(par, prg, DTYPE, DTYPE_real)

        self._weights = par["weights"]
        par["overlap"] = 1
        self._overlap = par["overlap"]
        self.par_slices = par["par_slices"]

        self.ratio = []
        for j in range(self.num_dev):
            self.ratio.append(
                clarray.to_device(
                    self.queue[4*j],
                    (par["weights"]).astype(
                        dtype=self.DTYPE_real)))

        self.unknown_shape = (self.NSlice, self.unknowns, self.dimY, self.dimX)
        self._grad_shape = self.unknown_shape + (4,)

        self._stream_grad = self._defineoperator(
            [self._grad],
            [self._grad_shape],
            [[self.unknown_shape]])

        self._stream_div = self._defineoperator(
            [self._div],
            [self.unknown_shape],
            [[self._grad_shape]],
            reverse_dir=True)

    def fwd(self, out, inp, **kwargs):
        """Forward operator application in-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event
            A PyOpenCL event to wait for.
        """
        self._stream_grad.eval(out, inp)

    def fwdoop(self, inp, **kwargs):
        """Forward operator application out-of-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        out = np.zeros(self._grad_shape, dtype=self.DTYPE)
        self._stream_grad.eval([out], inp)
        return out

    def adj(self, out, inp, **kwargs):
        """Adjoint operator application in-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex parameter space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex measurement space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        self._stream_div.eval(out, inp)

    def adjoop(self, inp, **kwargs):
        """Adjoint operator application out-of-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex measurement space which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        out = np.zeros(self.unknown_shape, dtype=self.DTYPE)
        self._stream_div.eval([out], inp)
        return out

    def _grad(self, outp, inp, par=None, idx=0, idxq=0,
              bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        return self.prg[idx].gradient(
            self.queue[4*idx+idxq],
            (self._overlap+self.par_slices, self.dimY, self.dimX),
            None, outp.data, inp[0].data,
            np.int32(self.unknowns),
            self.ratio[idx].data, self.DTYPE_real(self._dz),
            wait_for=outp.events + inp[0].events + wait_for)

    def _div(self, outp, inp, par=None, idx=0, idxq=0,
             bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        return self.prg[idx].divergence(
            self.queue[4*idx+idxq],
            (self._overlap+self.par_slices, self.dimY, self.dimX), None,
            outp.data, inp[0].data, np.int32(self.unknowns),
            self.ratio[idx].data, np.int32(bound_cond),
            self.DTYPE_real(self._dz),
            wait_for=outp.events + inp[0].events + wait_for)

    def getStreamedGradientObject(self):
        """Access privat stream gradient object.

        Returns
        -------
          PyqMRI.Streaming.Stream:
              A PyQMRI streaming object for the gradient computation.
        """
        return self._stream_grad


class OperatorFiniteSymGradientStreamed(Operator):
    """Streamed symmetrized gradient operator.

    This class implements the finite difference symmetrized gradient
    operation and the adjoint (negative symmetrized divergence).

    Parameters
    ----------
      par : dict A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      prg : PyOpenCL.Program
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator.
      DTYPE : numpy.dtype, numpy.complex64
        Complex working precission.
      DTYPE_real : numpy.dtype, numpy.float32
        Real working precission.

    Attributes
    ----------
      ctx : PyOpenCL.Context
        The context for the PyOpenCL computations.
      queue : PyOpenCL.Queue
        The computation Queue for the PyOpenCL kernels.
      par_slices : int
        Slices to parallel transfer to the compute device.
      ratio : list of PyOpenCL.Array
        Ratio between the different unknowns
    """

    def __init__(self, par, prg, DTYPE=np.complex64, DTYPE_real=np.float32):
        super().__init__(par, prg, DTYPE, DTYPE_real)

        par["overlap"] = 1
        self._overlap = par["overlap"]
        self.par_slices = par["par_slices"]

        unknown_shape = (self.NSlice, self.unknowns, self.dimY, self.dimX)
        self._grad_shape = unknown_shape + (4,)
        self._symgrad_shape = unknown_shape + (8,)

        self.ratio = []
        for j in range(self.num_dev):
            self.ratio.append(
                clarray.to_device(
                    self.queue[4*j],
                    (par["weights"]).astype(
                        dtype=self.DTYPE_real)))

        self._stream_symgrad = self._defineoperator(
            [self._symgrad],
            [self._symgrad_shape],
            [[self._grad_shape]],
            reverse_dir=True)

        self._stream_symdiv = self._defineoperator(
            [self._symdiv],
            [self._grad_shape],
            [[self._symgrad_shape]])

    def fwd(self, out, inp, **kwargs):
        """Forward operator application in-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event
            A PyOpenCL event to wait for.
        """
        self._stream_symgrad.eval(out, inp)

    def fwdoop(self, inp, **kwargs):
        """Forward operator application out-of-place.

        Apply the linear operator from parameter space to measurement space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex parameter space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        out = np.zeros(self._symgrad_shape, dtype=self.DTYPE)
        self._stream_symgrad.eval([out], inp)
        return out

    def adj(self, out, inp, **kwargs):
        """Adjoint operator application in-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array

        Parameters
        ----------
          out : PyOpenCL.Array
            The complex parameter space data which is the result of the
            computation.
          inp : PyOpenCL.Array
            The complex measurement space data which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        self._stream_symdiv.eval(out, inp)

    def adjoop(self, inp, **kwargs):
        """Adjoint operator application out-of-place.

        Apply the linear operator from measurement space to parameter space
        If streamed operations are used the PyOpenCL.Arrays are replaced
        by Numpy.Array
        This method need to generate a temporary array and will return it as
        the result.

        Parameters
        ----------
          inp : PyOpenCL.Array
            The complex measurement space which is used as input.
          wait_for : list of PyopenCL.Event
            A List of PyOpenCL events to wait for.

        Returns
        -------
          PyOpenCL.Array: A PyOpenCL array containing the result of the
          computation.
        """
        out = np.zeros(self._grad_shape, dtype=self.DTYPE)
        self._stream_symdiv.eval([out], inp)
        return out

    def _symgrad(self, outp, inp, par=None, idx=0, idxq=0,
                 bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        return self.prg[idx].sym_grad(
            self.queue[4*idx+idxq],
            (self._overlap+self.par_slices, self.dimY, self.dimX), None,
            outp.data, inp[0].data, np.int32(self.unknowns),
            self.ratio[idx].data,
            self.DTYPE_real(self._dz),
            wait_for=outp.events + inp[0].events + wait_for)

    def _symdiv(self, outp, inp, par=None, idx=0, idxq=0,
                bound_cond=0, wait_for=None):
        if wait_for is None:
            wait_for = []
        return self.prg[idx].sym_divergence(
            self.queue[4*idx+idxq],
            (self._overlap+self.par_slices, self.dimY, self.dimX), None,
            outp.data, inp[0].data,
            np.int32(self.unknowns),
            self.ratio[idx].data,
            np.int32(bound_cond),
            self.DTYPE_real(self._dz),
            wait_for=outp.events + inp[0].events + wait_for)

    def getStreamedSymGradientObject(self):
        """Access privat stream symmetrized gradient object.

        Returns
        -------
          PyqMRI.Streaming.Stream:
              A PyQMRI streaming object for the symmetrized gradient
              computation.
        """
        return self._stream_symgrad



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Module holding the class for streaming operations on the GPU."""

import numpy as np
import pyopencl as cl
import pyopencl.array as clarray


class Stream:
    """Basic streaming Class.

    This Class is responsible for performing asynchroneous transfer
    and computation on the GPU for arbitrary large numpy data.

    Parameters
    ----------
     fun : list of functions
        This list contains all functions that should be executed on the
        GPU. The functions are executed in order from first to last
        element of the list.
      outp_shape (list of tuple):
        The shape of the output array. Slice dimension is assumed to
        be the same as number of parallel slices plus overlap.
      inp_shape (list of list of tuple):
        The shape of the input arrays. Slice dimension is assumed to
        be the same as number of parallel slices plus overlap.
      par_slices : int
        Number of slices computed in one transfer on the GPU.
      overlap : int
        Overlap of adjacent blocks
      nslice : int
        Total number of slices
      queue : list of PyOpenCL.Queue
        The OpenCL queues used for transfer and computation. 4 queues are
        used per device.
      num_dev : int
        Number of computation devices.
      reverse : bool, false
        Indicator of the streaming direction. If False, streaming will
        start at the first and end at the last slice. If True streaming
        will be performed vice versa
      lhs : list of bool, None
        Indicator for the norm calculation in the line search of TGV.
        lhs refers to left hand side. Needs to be passed if a norm should
        be computed.
      DTYPE : numpy.dype, numpy.complex64
        Complex data type.

    Attributes
    ----------
      fun : list of functions
        This list contains all functions that should be executed on the GPU.
        The functions are executed in order from first to last element of the
        list.
      num_dev : int
        Number of computation devices.
      slices : int
        Number of slices computed in one transfer on the GPU.
      overlap : int
        Overlap of adjacent blocks
      queue : list of PyOpenCL.Queue
        The OpenCL queues used for transfer and computation. 4 queues are used
        per device.
      reverse : bool
        Indicator of the streaming direction. If False, streaming will start
        at the first and end at the last slice. If True streaming will be
        performed vice versa
      NSlice : int
        Total number of slices
      num_fun : int
        Total number of functions to stream (length of fun)
      lhs : list of bool, None
        Indicator for the norm calculation in the line search of TGV.
        lhs refers to left hand side.
      at_end : bool
        Specifies if the end of the data slice dimension is reached
      inp (list of list of list of PyOpenCL.Array):
        For each function a list of devices and a list of inputs is generated.
        E.g. for one function which needs two inputs and one computation device
        the list would have dimensions [1][1][2]
      outp (list of list of PyOpenCL.Array):
        For each function a list of devices with a single output is generated.
        E.g. for one function and one device
        the list would have dimension [1][1]
    """

    def __init__(self,
                 fun,
                 outp_shape,
                 inp_shape,
                 par_slices,
                 overlap,
                 nslice,
                 queue,
                 num_dev,
                 reverse=False,
                 lhs=None,
                 DTYPE=np.complex64):
        self.fun = fun
        self.num_dev = num_dev
        self.slices = par_slices
        self.overlap = overlap
        self.queue = queue
        self.reverse = reverse
        self.nslice = nslice
        self.num_fun = len(self.fun)
        self.dtype = DTYPE

        self.lhs = lhs
        self.at_end = False
        self.idx_todev_start = 0
        self.idx_todev_stop = 0
        self.idx_tohost_start = 0
        self.idx_tohost_stop = 0
        self._resetindex()

        self.inp = []
        self.outp = []

        self._alloctmparrays(inp_shape, outp_shape)

    def __add__(self, other):
        """Overloading add.

        Concatinates the functions, inputs and outputs of one stremaing
        object to another.

        Parameters
        ----------
          other (class stream):
            The object which should be added.

        Returns
        -------
          The combined objects
        """
        for j in range(other.num_fun):
            self.fun.append(other.fun[j])
            self.inp.append(other.inp[j])
            self.outp.append(other.outp[j])
        self.num_fun += other.num_fun
        return self

    def __del__(self):
        """Delete the Queue."""
        del self.queue

    def _alloctmparrays(self,
                        inp_shape,
                        outp_shape):
        block_size = self.slices+self.overlap
        for j in range(self.num_fun):
            self.inp.append([])
            for i in range(2*self.num_dev):
                self.inp[j].append([])
                for k in range(len(inp_shape[j])):
                    if not len(inp_shape[j][k]) == 0:
                        self.inp[j][i].append(
                            clarray.empty(
                                self.queue[4*int(i/2)],
                                ((block_size, )+inp_shape[j][k][1:]),
                                dtype=self.dtype))
                    else:
                        self.inp[j][i].append([])

        for j in range(self.num_fun):
            self.outp.append([])
            for i in range(2*self.num_dev):
                self.outp[j].append(
                    clarray.empty(
                        self.queue[4*int(i/2)],
                        ((block_size, )+outp_shape[j][1:]),
                        dtype=self.dtype))

    def _getindtodev(self):
        if self.reverse:
            tmp_return = slice(self.idx_todev_start,
                               self.idx_todev_stop)
            self.idx_todev_start -= self.slices
            self.idx_todev_stop -= self.slices
            if self.idx_todev_start < 0:
                self.idx_todev_start = 0
                self.idx_todev_stop = self.slices + self.overlap
        else:
            tmp_return = slice(self.idx_todev_start,
                               self.idx_todev_stop)
            self.idx_todev_start += self.slices
            self.idx_todev_stop += self.slices
            if self.idx_todev_stop > self.nslice:
                self.idx_todev_start = (self.nslice -
                                        self.slices -
                                        self.overlap)
                self.idx_todev_stop = self.nslice
        return tmp_return

    def _getindtohost(self):
        if self.reverse:
            tmp_return = slice(self.idx_tohost_start,
                               self.idx_tohost_stop)
            self.idx_tohost_start -= self.slices
            self.idx_tohost_stop -= self.slices
            self.at_end = False
            if self.idx_tohost_start < 0:
                self.at_end = True
                self.idx_tohost_start = 0
                self.idx_tohost_stop = self.slices + self.overlap
        else:
            tmp_return = slice(self.idx_tohost_start,
                               self.idx_tohost_stop)
            self.idx_tohost_start += self.slices
            self.idx_tohost_stop += self.slices
            self.at_end = False
            if self.idx_tohost_stop > self.nslice:
                self.at_end = True
                self.idx_tohost_start = (self.nslice -
                                         self.slices -
                                         self.overlap)
                self.idx_tohost_stop = self.nslice
        return tmp_return

    def eval(self, outp, inp, par=None):
        """Evaluate all functions of the object.

        Perform asynchroneous evaluation of the functions stored in
        fun.

        Parameters
        ----------
          outp (list of np.arrays):
            Result of the computation for each function as numpy array
          inp (list of list of np.arrays):
            For each function contains a list of numpy arrays used as input.
          par (list of list of parameters):
            Optional list of parameters which should be passed to a function.
        """
        # Reset Array Index
        self._resetindex()
        # Warmup Queue 1
        self._streamtodevice(inp, 0)
        self._startcomputation(par, bound_cond=1, odd=0)

        # Warmup Queue 1
        self._streamtodevice(inp, 1)
        self._startcomputation(par, bound_cond=0, odd=1)

        # Start Streaming
        islice = 2*self.slices*self.num_dev
        odd = True
        while islice+self.overlap < self.nslice:
            # Collect Previous Block
            odd = not odd
            self._streamtohost(outp, odd)
            # Stream new Block
            self._streamtodevice(inp, odd)
            # Start Computation
            self._startcomputation(par, bound_cond=0, odd=odd)
            islice += self.num_dev*self.slices

        # Collect last block
        if odd:
            self._streamtohost(outp, 0)
            self._streamtohost(outp, 1)
        else:
            self._streamtohost(outp, 1)
            self._streamtohost(outp, 0)
        # Wait for all Queues to finish
        for i in range(self.num_dev):
            self.queue[4*i].finish()
            self.queue[4*i+1].finish()
            self.queue[4*i+2].finish()
            self.queue[4*i+3].finish()

    def evalwithnorm(self, outp, inp, par=None):
        """Evaluate all functions of the object and returns norms.

        Perform asynchroneous evaluation of the functions stored in
        fun. Same as eval but also computes the norm relevant for the
        linesearch in the TGV algorithm.

        Parameters
        ----------
          outp : list of np.arrays
            Result of the computation for each function as numpy array
          inp (list of list of np.arrays):
            For each function contains a list of numpy arrays used as input.
          par : list of list of parameters
            Optional list of parameters which should be passed to a function.

        Returns
        -------
          tuple of floats:
            (lhs, rhs) The lhs and rhs for the linesearch used in the TGV
            algorithm.
        """
        # Reset Array Index
        self._resetindex()
        rhs = 0
        lhs = 0
        # Warmup Queue 1
        self._streamtodevice(inp, 0)
        self._startcomputation(par, bound_cond=1, odd=0)

        # Warmup Queue 2
        self._streamtodevice(inp, 1)
        self._startcomputation(par, bound_cond=0, odd=1)

        # Start Streaming
        islice = 2*self.slices*self.num_dev
        odd = True
        while islice + self.overlap < self.nslice:
            odd = not odd
            # Collect Previous Block
            (rhs, lhs) = self._streamtohostnorm(
                outp,
                rhs,
                lhs,
                odd)
            # Stream new Block
            self._streamtodevice(inp, odd)
            # Start Computation
            islice += self.num_dev*self.slices
            self._startcomputation(par, bound_cond=0, odd=odd)

        # Collect last block
        if odd:
            (rhs, lhs) = self._streamtohostnorm(outp, rhs, lhs, 0)
            (rhs, lhs) = self._streamtohostnorm(outp, rhs, lhs, 1)
        else:
            (rhs, lhs) = self._streamtohostnorm(outp, rhs, lhs, 1)
            (rhs, lhs) = self._streamtohostnorm(outp, rhs, lhs, 0)
        # Wait for all Queues to finish
        for i in range(self.num_dev):
            self.queue[4*i].finish()
            self.queue[4*i+1].finish()
            self.queue[4*i+2].finish()
            self.queue[4*i+3].finish()
        return (lhs, rhs)

    def _streamtodevice(self, inp, odd):
        for idev in range(self.num_dev):
            idx = self._getindtodev()
            for ifun in range(self.num_fun):
                if not len(inp[ifun]) == 0:
                    for iinp in range(len(self.inp[ifun][idev])):
                        if not len(inp[ifun][iinp]) == 0:
                            self.inp[
                                ifun][
                                    2*idev+odd][
                                        iinp].add_event(
                                            cl.enqueue_copy(
                                                self.queue[4*idev+odd],
                                                self.inp[
                                                    ifun][
                                                        2*idev+odd][
                                                            iinp].data,
                                                inp[
                                                    ifun][
                                                        iinp][idx, ...],
                                                wait_for=self.inp[
                                                    ifun][
                                                        2*idev+odd][
                                                            iinp].events,
                                                is_blocking=False))
                            self.queue[4*idev+odd].flush()

    def _startcomputation(self, par=None, bound_cond=0, odd=0):
        if par is None:
            par = []
            for ifun in range(self.num_fun):
                par.append([])
        for idev in range(self.num_dev):
            for ifun in range(self.num_fun):
                for inps in self.inp[ifun][2*idev+odd]:
                    for inp in inps:
                        for event in inp.events:
                            event.wait()
                self.outp[
                    ifun][
                        2*idev+odd].add_event(
                            self.fun[ifun](
                                self.outp[ifun][2*idev+odd],
                                self.inp[ifun][2*idev+odd][:],
                                par[ifun],
                                idev,
                                odd,
                                bound_cond=bound_cond))
                self.queue[4*idev+odd].flush()
            bound_cond = 0

    def _streamtohost(self, outp, odd):
        for idev in range(self.num_dev):
            self.queue[4*idev+3-odd].finish()
            if self.num_dev > 1:
                self.queue[4*np.mod(idev-1, self.num_dev)+2+odd].finish()
                self.queue[4*np.mod(idev-1, self.num_dev)+3-odd].finish()
            idx = self._getindtohost()
            for ifun in range(self.num_fun):
                self.outp[ifun][2*idev+odd].add_event(
                    cl.enqueue_copy(
                        self.queue[4*idev+2+odd],
                        outp[ifun][idx, ...],
                        self.outp[ifun][2*idev+odd].data,
                        wait_for=self.outp[ifun][2*idev+odd].events,
                        is_blocking=False))
                self.queue[4*idev+2+odd].flush()

    def _streamtohostnorm(self, outp, rhs, lhs, odd):
        for idev in range(self.num_dev):
            self.queue[4*idev+3-odd].finish()
            if self.num_dev > 1:
                self.queue[4*np.mod(idev-1, self.num_dev)+2+odd].finish()
                self.queue[4*np.mod(idev-1, self.num_dev)+3-odd].finish()
            idx = self._getindtohost()
            for ifun in range(self.num_fun):
                self.outp[ifun][2*idev+odd].add_event(
                    cl.enqueue_copy(
                        self.queue[4*idev+2+odd],
                        outp[ifun][idx, ...],
                        self.outp[ifun][2*idev+odd].data,
                        wait_for=self.outp[ifun][2*idev+odd].events,
                        is_blocking=False))
                if self.reverse:
                    if not self.at_end:
                        (rhs, lhs) = self._calcnormreverse(
                            rhs, lhs, idev, ifun, odd)
                    else:
                        (rhs, lhs) = self._calcnormforward(
                            rhs, lhs, idev, ifun, odd)
                else:
                    if not self.at_end:
                        (rhs, lhs) = self._calcnormforward(
                            rhs, lhs, idev, ifun, odd)
                    else:
                        (rhs, lhs) = self._calcnormreverse(
                            rhs, lhs, idev, ifun, odd)
                self.queue[4*idev+odd].flush()
                self.queue[4*idev+2+odd].flush()
        return (rhs, lhs)

    def _resetindex(self):
        if self.reverse:
            self.idx_todev_start = self.nslice - (self.slices + self.overlap)
            self.idx_todev_stop = self.nslice
            self.idx_tohost_start = self.nslice - (self.slices +
                                                   self.overlap)
            self.idx_tohost_stop = self.nslice
        else:
            self.idx_todev_start = 0
            self.idx_todev_stop = (self.slices + self.overlap)
            self.idx_tohost_start = 0
            self.idx_tohost_stop = (self.slices + self.overlap)

    def connectouttoin(self, outpos, inpos):
        """Connect output to input of functions within the object.

        This function can be used to connect the output of a function to the
        input of another one used in the same stream object.

        Parameters
        ----------
          outpos : int
            The position in the list of outputs which should be connected to
            an input
          inpos (list of list of np.arrays:
            The position in the list of inputs which should be connected
            with an output
        """
        for j in range(2*self.num_dev):
            self.inp[inpos[0]][j][inpos[1]] = self.outp[outpos][j]

    def _calcnormreverse(self, rhs, lhs, idev, ifun, odd=0):
        if self.lhs[ifun] is False:
            rhs += clarray.vdot(
                self.outp[
                    ifun][
                        2*idev+odd][self.overlap:, ...] -
                self.inp[
                    ifun][
                        2*idev+odd][0][self.overlap:, ...],
                self.outp[
                    ifun][
                        2*idev+odd][self.overlap:, ...] -
                self.inp[
                    ifun][
                        2*idev+odd][0][self.overlap:, ...]
                ).get()
        else:
            lhs += clarray.vdot(
                self.outp[
                    ifun][
                        2*idev+odd][self.overlap:, ...] -
                self.inp[
                    ifun][
                        2*idev+odd][-1][self.overlap:, ...],
                self.outp[
                    ifun][
                        2*idev+odd][self.overlap:, ...] -
                self.inp[
                    ifun][
                        2*idev+odd][-1][self.overlap:, ...]
                ).get()
        return (rhs, lhs)

    def _calcnormforward(self, rhs, lhs, idev, ifun, odd=0):
        if self.lhs[ifun] is False:
            rhs += clarray.vdot(
                self.outp[
                    ifun][
                        2*idev+odd][:self.slices, ...] -
                self.inp[
                    ifun][
                        2*idev+odd][0][:self.slices, ...],
                self.outp[
                    ifun][
                        2*idev+odd][:self.slices, ...] -
                self.inp[
                    ifun][
                        2*idev+odd][0][:self.slices, ...]
                ).get()
        else:
            lhs += clarray.vdot(
                self.outp[
                    ifun][
                        2*idev+odd][:self.slices, ...] -
                self.inp[
                    ifun][
                        2*idev+odd][-1][:self.slices, ...],
                self.outp[
                    ifun][
                        2*idev+odd][:self.slices, ...] -
                self.inp[
                    ifun][
                        2*idev+odd][-1][:self.slices, ...]
                ).get()
        return (rhs, lhs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Utility functions for PyQMRI fitting."""
import configparser
import os
from pyqmri.transforms import PyOpenCLnuFFT


def prime_factors(n):
    """Prime factorication.

    Parameters
    ----------
      n : int
        Value which should be factorized into primes.
    """
    i = 2
    factors = []
    while i * i <= n:
        if n % i:
            i += 1
        else:
            n //= i
            factors.append(i)
    if n > 1:
        factors.append(n)
    return factors


def NUFFT(par, trafo=True, SMS=False):
    """NUFFT for image guess.

    Parameters
    ----------
      par : dict
        Parameter struct to setup the NUFFT
      trafo : bool, True
        Radial (True) or Cartesian (False) FFT.
      SMS : bool, False
        SMS (True) or normal FFT (False, default).
    """
    NC = par["NC"]
    NScan = par["NScan"]
    par["NC"] = 1
    par["NScan"] = 1

    FFT = (PyOpenCLnuFFT.create(
        par["ctx"][0], par["queue"][0], par,
        radial=trafo, SMS=SMS))
    par["NC"] = NC
    par["NScan"] = NScan

    return FFT


def gen_default_config():
    """Generate default config file."""
    config = configparser.ConfigParser()

    config['TGV'] = {}
    config['TGV']["max_iters"] = '1000'
    config['TGV']["start_iters"] = '10'
    config['TGV']["max_gn_it"] = '7'
    config['TGV']["lambd"] = '1e2'
    config['TGV']["gamma"] = '5e-2'
    config['TGV']["delta"] = '1e-1'
    config['TGV']["omega"] = '0'
    config['TGV']["display_iterations"] = '0'
    config['TGV']["gamma_min"] = '5e-3'
    config['TGV']["delta_max"] = '1e2'
    config['TGV']["omega_min"] = '0'
    config['TGV']["tol"] = '1e-6'
    config['TGV']["stag"] = '1e10'
    config['TGV']["delta_inc"] = '10'
    config['TGV']["gamma_dec"] = '0.5'
    config['TGV']["omega_dec"] = '0.5'

    config['TV'] = {}
    config['TV']["max_iters"] = '1000'
    config['TV']["start_iters"] = '10'
    config['TV']["max_gn_it"] = '7'
    config['TV']["lambd"] = '1e2'
    config['TV']["gamma"] = '5e-2'
    config['TV']["delta"] = '1e-1'
    config['TV']["omega"] = '0'
    config['TV']["display_iterations"] = '0'
    config['TV']["gamma_min"] = '5e-3'
    config['TV']["delta_max"] = '1e2'
    config['TV']["omega_min"] = '0'
    config['TV']["tol"] = '1e-6'
    config['TV']["stag"] = '1e10'
    config['TV']["delta_inc"] = '10'
    config['TV']["gamma_dec"] = '0.5'
    config['TV']["omega_dec"] = '0.5'

    with open('default.ini', 'w') as configfile:
        config.write(configfile)


def read_config(conf_file, reg_type="TGV"):
    """Config file reader.

    Parameters
    ----------
      conf_file : str
        Path to config file
      reg_type : str, TGV
        Select witch regularization parameters from the file should be used.
    """
    config = configparser.ConfigParser()

    if not conf_file.endswith('.ini'):
        conf_file += '.ini'
    try:
        with open(conf_file, 'r') as f:
            config.read_file(f)
    except BaseException:
        print("Config file not readable or not found. "
              "Falling back to default.")
        gen_default_config()
        with open('default.ini', 'r') as f:
            config.read_file(f)
    finally:
        params = {}
        for key in config[reg_type]:
            if key in {'max_gn_it', 'max_iters', 'start_iters'}:
                params[key] = int(config[reg_type][key])
            elif key == 'display_iterations':
                params[key] = config[reg_type].getboolean(key)
            else:
                params[key] = float(config[reg_type][key])
        return params


def save_config(conf, path, reg_type="TGV"):
    """Config file writer.

    Save the used config alongside the results.

    Parameters
    ----------
      conf : str
        Path to config file
      path : str
        Output path
      reg_type : str, TGV
        Select witch regularization parameters from the file should be used.
    """
    tmp_dict = {}
    tmp_dict[reg_type] = conf
    config = configparser.ConfigParser()
    config.read_dict(tmp_dict)
    with open(path+os.sep+'config.ini', 'w') as configfile:
        config.write(configfile)
